#
# Copyright 2005, Shane Stephens.
#
# This work is protected by the BSD license.  Please see license.txt in the
# root directory for more information.
#

copyright_notice = [
"/*\n",
" * Copyright 2005, Shane Stephens.\n",
" *\n",
" * This work is protected by the BSD license.  Please see license.txt in\n",
" * the root directory for more information.\n",
" */\n"
]

from dictparser import *
import sys

class UnknownTypeException(TipException):
	def __init__(this, name):
		this.string = "I don't know how to handle type %s" % (name,)

class BadArraySpecificationException(TipException):
	def __init__(this, name):
		this.string = "Badly specified array: " + name

class BadStringIndentException(TipException):
	def __init__(this):
		this.string = "Indented to less than 0"

class MultipleArrayAccessorException(TipException):
	def __init__(this, name):
		this.string = "class " + name + " has multiple primary setters"\
			" or getters"

class EmptyBodyException(TipException):
	def __init__(this, name):
		this.string = "function " + name + " is not external and has "\
		"no function body defined"

class CodeString:
	def __init__(this, initial=""):
		this.ind = 0
		this.string = initial
		this.ended_on_newline = len(initial) > 0 and initial[-1] == "\n"
	def __iadd__(this, other):
		try:
			other2 = other.string
			other = other2
		except:
			pass

		if this.ended_on_newline:
			this.string += "\t" * this.ind
			this.ended_on_newline = 0
		while len(other) > 0:
			try:
				index = other.index("\n")
				this.string += other[:index]
				this.string += "\n"
				other = other[index+1:]
				if len(other) > 0:
					this.string += "\t" * this.ind
				else:
					this.ended_on_newline = 1
			except:
				this.string += other
				other = ""
		return this
	def indent(this):
		this.ind += 1
	def outdent(this):
		this.ind -= 1
		if this.ind < 0:
			raise BadStringIndentException()
	def append_to(this, file):
		f = open(file, 'a')
		f.write(this.string)
		f.close()
	def write_to(this, file):
		f = open(file, 'w')
		f.write(this.string)
		f.close()
		

#
# Class to represent an interface.  It contains a list of functions and a
# name.  Maybe later it will have more.
#
class TipInterface:
	def __init__(this, props):
		this.name = props["name"]
		this.props = {"strictChecking":"0", "logAccesses":"0"}
		this.props.update(props)
		this.functions = []
		
	def add_function(this, function):
		this.functions.append(function)

	def __str__(this):
		s = this.name + "\n"
		for p in this.functions:
			s += "\t" + str(p)
		return s
	
	#
	# generate the defining struct for this interface.
	# This is for interface.h
	#
	def struct_definition(this):
		s = CodeString("typedef struct {\n")
		s.indent()
		for function in this.functions:
			s += function.prototype_pointer()
		s.outdent()
		s += "} %s;\n\n" % (this.name,)
		return s

	#
	# generate an individual class' implementation data for this interface.
	# this goes at the top of each .c file
	#
	def definitions(this, className):
		s = CodeString()
		for function in this.functions:
			s += function.definition(className)
		return s

	def implementation(this, className):
		s = CodeString("%s %s_%s = {" % (this.name, className, 
			this.name))
		s += reduce(lambda a, b: a + ", " + b, map(
			lambda a: "C_" + className + "_" + a.name,
			this.functions))
		s += "};\n\n"
		return s

	#
	# generate this interface's accessor macros
	#
	def accessor_defines(this):
		if this.props["strictChecking"] == "1":
			s = CodeString("static inline %s *%s_IF(void *x) {\n" %
				(this.name, this.name.upper()))
			s.indent()
			s += "assert (((TograMin *)x)->interfaces->%s" % \
				(this.name,) + " != NULL);\n"
			if this.props["logAccesses"] == "1":
				s += 'printf("%s' % (this.name,) +\
					'access for %p (%p)\\n", x, ' +\
					'((TograMin *)x)->interfaces->%s);\n' %\
					(this.name,)
			
			s += "return ((TograMin *)x)->interfaces->%s;\n" % \
				(this.name,)
			s.outdent()
			s += "}\n"
		else:
			s = CodeString("#define %s_IF(x) \\\n" % \
				(this.name.upper()))
			s.indent()
			s += "(((TograMin *)x)->interfaces->%s)\n" % (this.name)
			s.outdent()
		s += "#define IS_%s(x) \\\n" % (this.name.upper())
		s.indent()
		s += "(%s_IF(x) != NULL)\n" % (this.name.upper())
		s.outdent()
		
		for function in this.functions:
			s += "#define %s_%s(" % \
				(this.name.upper(), function.name.upper())
			flist = reduce(lambda a, b: a + ", " + b.name,
				function.parameters, "x")
			if function.returnType is not None:
				flist += ", rt"
			s += flist
			s += ") \\\n"
			s.indent()
			s += "(%s_IF(x)->%sFunc(%s))\n" % \
				(this.name.upper(), function.name, flist)
			s.outdent()
		s += "\n"
		return s
	

conversion_lookup = {'int': 'return PyInt_FromLong((long)return_value)',
	'PyObject *': 'return return_value'}

#
# This class defines both a function prototype and an individual implementation
# of that prototype
#
class TipFunction:
	def __init__(this, props):
		this.name = props["name"]
		this.values = {"external":0}
		this.values.update(props)
		this.parameters = []
		this.returnType = None
		this.text = []
	
	#
	# Add a parameter to the function
	#
	def add_parameter(this, parameter):
		this.parameters.append(parameter)

	#
	# generate the 'return string' for the function.  All
	# functions should return TRUE if they worked, FALSE otherwise,
	# so this actually goes at the end!
	#
	def return_str(this):
		if this.returnType == None:
			return ""
		if this.returnType[0] == "I":
			return ", PyObject **return_value"
		return ", %s *return_value" % (this.returnType,)
	
	def set_text(this, text):
		this.text += text
	
	def __str__(this):
		s = this.name + ": "
		for p in this.parameters:
			s += str(p) + ", "
		return s[:-2] + "\n"
	
	#
	# Generate a pointer declaration for the prototype
	#
	def prototype_pointer(this):
		s = CodeString("int (*%sFunc)(" % (this.name,))
		s += this.input_definition()
		s += this.return_str()
		s += ");\n"
		return s

	#
	# Generate an input definition for the prototype
	#
	def input_definition(this):
		s = CodeString("PyObject *self")
		for p in this.parameters:
			if p.type[0] == "I":
				s += ", PyObject * %s" % (p.name,)
			else:
				s += ", %s %s" % (p.type, p.name)
		return s
		
	#
	# Generate common "head" code for definition and declaration
	#
	def head(this, className):
		s = CodeString("int C_%s_%s(" % (className, this.name))
		s += this.input_definition()
		s += this.return_str()
		s += ")"
		return s

	#
	# Generate a function definition
	#
	def definition(this, className):
		s = this.head(className)
		s += ";\n"
		return s
	
	#
	# Generate the function, wrapped to be accessible in python
	#
	def generate_function(this, className):
		s = CodeString(
			"PyObject *%s_%s(PyObject *self, PyObject *args) {\n" %
				(className, this.name))
		s.indent()
		for item in this.parameters:
			s += item.declaration(0)
		s += "\n"

		if len(this.parameters) > 0:
			s += 'if (!PyArg_ParseTuple(args, "'
			s += reduce(lambda a, b: a + b,
				map(lambda a: a.parse_tuple_format(),
					this.parameters))
			s += '"'
			for item in this.parameters:
				s += ', '
				s += item.parse_tuple_inputs()
			s += ")) {\n"
			s.indent()
			s += "return NULL;\n"
			s.outdent()
			s += "}\n\n"

		convert_items = filter(lambda a: 
			a.values["pythonType"] is not None,
			this.parameters)

		for item in convert_items:
			s += item.conversion_code()

		if not this.returnType is None:
			s += "%s return_value;\n" % (this.returnType)
			
		s += "%s *object = (%s *)self;\n\n" % (className, className)

		if this.values["external"] == "1":
			s += "if(!"
			s += "C_%s_%s(" % (className, this.name)
			s += reduce(lambda a, b: a + ", " + b.name, 
				this.parameters, "object")
			if not this.returnType is None:
				s += ", &return_value"
			s += "))\n"
			s.indent()
			s += "return NULL;\n"
			s.outdent()
		
			s2 = CodeString()
			s2 += "int "
			s2 += "C_%s_%s(" % (className, this.name)
			s2 += reduce(
				lambda a, b: "%s, %s %s" % 
				(a, b.realtype(), b.name),
				this.parameters, "%s *object" % (className,))
			if not this.returnType is None:
				s2 += ", %s *return_value" % (this.returnType)
			s2 += ");\n"
		else:
			s2 = CodeString()
			if len(this.text) == 0:
				raise EmptyBodyException(this.name)
			s += reduce(lambda a, b: a + "\n" + b, this.text)
			s2 += this.definition(className) 
			s2 += "\n\n"
		s += "\n\n"
		
		s2.append_to('%s_common.h' % (className,))

		if this.returnType is None:
			s += "Py_INCREF(Py_None);\nreturn Py_None;\n"
		else:
			s += conversion_lookup[this.returnType] + ";\n"
		
		s.outdent()
		s += "}\n"
		return s
	
	def generate_method_def(this, className):
		return CodeString(
			"{\"%s\", (PyCFunction)%s_%s, METH_VARARGS, \"\"},\n"
			%
			(this.name, className, this.name)
		)

class TipExtern:
	def __init__(this, name, type):
		this.type = type
		this.name = name
	def declaration(this):
		return "%s %s;\n" % (this.type, this.name)

class TipDefineGroup:
	def __init__(this):
		this.items = []
	def add_item(this, name):
		this.items.append(name)
	def define_directive(this):
		s = CodeString()
		for item in range(len(this.items)):
			s += "#define %s %d\n" % (this.items[item], item)
		return s

class TipClass:
	def __init__(this, props):
		this.name = props["name"]
		try:
			this.pythonName = props["pythonName"]
		except:
			this.pythonName = this.name[2:]
		this.interfaces = []
		this.text = []
		this.functions = []
		this.items = []
		this.init_inputs = []
		this.text = []
		this.free_text = []
		this.externs = []
		this.define_groups = []
	def add_interface(this, interface):
		this.interfaces.append(interface)
	def add_function(this, function):
		this.functions.append(function)
	def add_item(this, item):
		this.items.append(item)
	def add_init_input(this, item):
		this.init_inputs.append(item)
	def add_extern(this, extern):
		this.externs.append(extern)
	def add_define_group(this, group):
		this.define_groups.append(group)
	def set_init_text(this, text):
		this.text += text
	def set_free_text(this, text):
		this.free_text += text

	#
	# Generate new function definition
	#
	def generate_new_definition(this):
		s = CodeString("\nPyObject *C_%s_New(" % (this.name,))
		on_init_items = filter(lambda a: a.values["onInit"], this.items)
		
		if len(this.init_inputs + on_init_items) > 0:
			s += reduce(lambda a, b: a + ", " + b,
				map(lambda a: a.get_calling_arguments(),
				this.init_inputs + on_init_items))
		else:
			s += "void"
		
		s += ")"
		return s
	
	#
	# Generate a methodDef entry for Togra.c
	#
	def generate_methodDef(this):
		s = CodeString('{"%s", (PyCFunction)%s_New, ' % 
			(this.pythonName, this.name))
		s += "METH_VARARGS"
		
		on_init_items = filter(lambda a: a.values["onInit"], this.items)
		keyword_inputs = filter(lambda a: a.values["keyword"], 
			this.init_inputs + on_init_items)
		if len(keyword_inputs) > 0:
			s += " | METH_KEYWORDS"
		s += ', ""},\n'
		return s
		
	
	#
	# Generate the code for this class, both struct definition in
	# interface.h and C code in classname_impl.c
	#
	def generate_class_code(this, all_interfaces):
		#
		# First, dump my struct to the interface file
		#
		s = CodeString()
		forward_defs = filter(
			lambda a: a.values["definition"] == "forward",
			this.items)
		for forward_def in forward_defs:
			name = forward_def.type
			if name[-1] == "*":
				name = name[:-2]
			s += "typedef struct %s %s;\n" % (name, name)
		s += "\n"
			
		s += "typedef struct {\n"
		s.indent()
		s += "PyObject_HEAD\n"
		s += "IC *interfaces;\n"
		for item in this.items:
			s += item.declaration(1)
		s.outdent()
		s += "} %s;\n\n" % (this.name,)
		s.append_to('interfaces.h')

		#
		# Now, dump my function defs to the definitions file
		#
		s = this.generate_new_definition()
		s += ";\n"
		s.append_to('definitions.h')

		#
		# Now build a file that both I and the manual code will include
		#
		s = CodeString()
		
		#
		# generate any defines
		#
		for group in this.define_groups:
			s += group.define_directive()
			s += "\n"

		#
		# Declare any extern variables
		#
		for extern in this.externs:
			s += "extern " + extern.declaration()
		

		for interface in this.interfaces:
			s += interface.definitions(this.name)

		#
		# Now, dump my <name>_impl.c file
		# Start with the interface definition
		#
		s.write_to('%s_common.h' % (this.name))
		
		s = CodeString()
		for line in copyright_notice:
			s += line

		s += '#include "TograIncludes.h"\n'
		s += '#include "%s_common.h"\n\n' % (this.name)
		s += "\n"
		for interface in this.interfaces:
			s += interface.implementation(this.name)
		s += "staticforward PyTypeObject %sType;\n\n" % (this.name,)
		s += "static IC %s_Interfaces = {\n" % (this.name,)
		s.indent()
		implemented_interfaces = map(lambda a: a.name, this.interfaces)
		for interface in all_interfaces:
			if interface in implemented_interfaces:
				s += "&%s_%s,\n" % (this.name, interface)
			else:
				s += "NULL,\n"
		s.outdent()
		s += "};\n\n"

		#
		# Next, build the new function (C implementation)
		#
		s += this.generate_new_definition()
		s += " {\n\n"
		s.indent()

		s += "%s *object = (%s *)PyObject_New(%s, &%sType);\n\n" % \
			(this.name, this.name, this.name, this.name)
		s += "object->interfaces = &%s_Interfaces;\n\n" % \
			(this.name,)
		# write the user-defined text BEFORE the majority of
		# initialisation, in case they want to do things like
		# allocate structs
		for line in this.text:
			s += line + "\n"
		s += "\n"

		on_init_items = filter(lambda a: a.values["onInit"], this.items)
		for item in on_init_items:
			s += item.variable_initialisation()
		
		constant_init_items = filter(lambda a: 
			a.values["default"] is not None and 
			not a.is_optional(), this.items)
		for item in constant_init_items:
			s += item.constant_initialisation()
		
		s += "\n"
		
		#
		# increment references for PyObjects
		#
		managed_items = filter(lambda a: a.needs_managing(), 
			on_init_items)

		for item in managed_items:
			s += item.management_code(item.name)
		
		s += "\nreturn (PyObject *)object;\n"
		s.outdent()
		s += "}\n\n"
		
		#
		# Next build the python front-end for the new function
		#
		s2 = CodeString()
		s2 += "PyObject *%s_New(PyObject *self, PyObject *args" %\
			(this.name,)

		keyword_inputs = filter(lambda a: a.values["keyword"], 
			this.init_inputs + on_init_items)
		if len(keyword_inputs) > 0:
			s2 += ", PyObject *kwords"
		
		s2 += ")"
		s += s2
		
		# write the front-end into Togra.c
		s2 += ";\n"
		s2.append_to("Togra.c")
		
		s += "{\n\n"
		s.indent()
		# variable definitions

		for item in this.init_inputs + on_init_items:
			s += item.declaration(0)

		s += "\n"

		if (len(this.init_inputs + on_init_items) > 0):
			if len(keyword_inputs) > 0:
				unkeyword_inputs = filter(
					lambda a: a not in keyword_inputs,
					this.init_inputs + on_init_items)
		
				opt_inputs = filter(
					lambda a: a.is_optional(),
					unkeyword_inputs)
				unopt_inputs = filter(
					lambda a: not a.is_optional(),
					unkeyword_inputs)
		
				# keyword list
				s += "static char *kwlist[] = {"
				for input in unkeyword_inputs + keyword_inputs:
					s += '"%s", ' % (input.name,)
				s += "NULL};\n\n"
			
				# PyArg_ParseTupleAndKeywords code
				s += "if (!PyArg_ParseTupleAndKeywords("
				s += 'args, kwords, "'
				if len(unopt_inputs) > 0:
					s += reduce(lambda a, b : a + b, map(
						lambda a: 
							a.parse_tuple_format(), 
							unopt_inputs))
				s += "|"
				if len(opt_inputs) > 0:
					s += reduce(lambda a, b : a + b, map(
						lambda a: 
							a.parse_tuple_format(), 
							opt_inputs))
				s += reduce(lambda a, b : a + b, map(
					lambda a: a.parse_tuple_format(), 
						keyword_inputs))
				s += '", kwlist'

			else:
				# PyArg_ParseTuple code
				s += 'if (!PyArg_ParseTuple(args, "'
				opt_inputs = filter(
					lambda a: a.is_optional(),
					this.init_inputs + on_init_items)
				unopt_inputs = filter(
					lambda a: not a.is_optional(),
					this.init_inputs + on_init_items)
				if len(unopt_inputs) > 0:
					s += reduce(lambda a, b : a + b, map(
						lambda a: 
						a.parse_tuple_format(), 
						unopt_inputs))
				if len(opt_inputs) > 0:
					s += "|"
					s += reduce(lambda a, b: a + b, map(
						lambda a: 
						a.parse_tuple_format(),
						opt_inputs))
				s += '"'
			
			for input in (unopt_inputs + opt_inputs + 
				keyword_inputs):
				s += ", "
				s += input.parse_tuple_inputs()
			s += ")) {\n"
			s.indent()
			s += "return NULL;\n"
			s.outdent()
			s += "}\n\n"

			managed_kis = filter(lambda a: a.needs_managing(), 
				keyword_inputs)
			if len(managed_kis) > 0:
				for mki in managed_kis:
					s += mki.set_nokeyword_value()

		# conversion (from PythonType to CType)
		convert_items = filter(lambda a: 
			a.values["pythonType"] is not None,
			this.init_inputs + on_init_items)
		
		for item in convert_items:
			s += item.conversion_code()

	
		# Callthrough to C
		s += "return (PyObject *)C_%s_New(" % (this.name,)
		s += reduce(lambda a, b: a + ", " + b.getCallArguments(), 
			this.init_inputs + on_init_items, "")[2:]
		s += ");\n"
		s.outdent()
		s += "}\n\n"

		#
		# Now build item setters and getters
		#
		for input in this.items:
			s += input.generate_setget(this.name)
		
		#
		# Now build the functions
		#
		for function in this.functions:
			s += function.generate_function(this.name)
			s += "\n"

		#
		# Now build the methods struct for the functions
		#

		s += "static PyMethodDef %s_methods[] = {\n" % (this.name,)
		s.indent()
	
		for input in this.items:
			s += input.generate_setget_record(this.name)
			

		for function in this.functions:
			s += function.generate_method_def(this.name)
		
		s += "{NULL, NULL, 0, NULL}\n"
		s.outdent()
		s += "};\n\n"

		s += "static PyObject *%s_GetAttr(" % (this.name,)
		s += "PyObject *self, char *name) {\n"
		s.indent()
		s += "return Py_FindMethod(%s_methods, self, name);\n" % \
			(this.name,)
		s.outdent()
		s += "}\n\n"
	
		#
		# Now build the free function
		#
		s += "static void %s_Free(PyObject *self) {\n" % (this.name,)
		s.indent()
		s += "%s *object = (%s *)self;\n" % (this.name, this.name)
		for line in this.free_text:
			s += line
			s += "\n"

		#
		# decrement references for PyObjects
		#
		managed_items = filter(lambda a: a.needs_managing(), 
			on_init_items)

		for item in managed_items:
			s += item.unmanagement_code("object->" + item.name)
			
		s += "PyObject_Del(object);"
			
		s.outdent()
		s += "}\n\n"
		
		#
		# Make sequence methods
		#
		setters = filter(lambda a: a.is_primary_setter(), this.items)
		getters = filter(lambda a: a.is_primary_getter(), this.items)
	
		if len(setters) > 1 or len(getters) > 1:
			raise MultipleArrayAccessorException(this.name)
		if len(setters) == 1 or len(getters) == 1:
			s += "static PySequenceMethods "
			s += "%s_sequenceMethods = {\n"  % (this.name,)
			s.indent()
			s += "%s_Length,\n" % (this.name,)
			s += "0,\n"
			s += "0,\n"
			if len(getters) == 1:
				s += "%s_Item,\n" % (this.name,)
			else:
				s += "0,\n"
			s += "0,\n"
			if len(setters) == 1:
				s += "%s_AssItem,\n" % (this.name,)
			else:
				s += "0,\n"
			s += "0,\n"
			s += "0,\n"
			s += "0,\n"
			s += "0\n"
			s.outdent()
			s += "};\n\n"
					
		#
		# Finally, build the type structure
		#
		s += "static PyTypeObject %sType = {\n" % (this.name,)
		s.indent()
		s += "PyObject_HEAD_INIT(NULL)\n"
		s += "0,\n"
		s += "\"%s\",\n" % (this.name,)
		s += "sizeof(%s),\n" % (this.name,)
		s += "0,\n"
		s += "%s_Free,\n" % (this.name,)
		s += "0,\n"
		s += "%s_GetAttr,\n" % (this.name)
		s += "0,\n"
		s += "0,\n"
		s += "0,\n"
		s += "0,\n"

		if len(setters) == 1 or len(getters) == 1:
			s += "&%s_sequenceMethods,\n" % (this.name,)
	
		s.outdent()
		s += "};\n"

		s.write_to(this.name + '_impl.c')

#
# What's the CType for the pythonType?
#
pt_lookup = {
	'PyString' : 'PyObject *'
		}
#
# What's the argstring entry for the pythonType?
#
pae_lookup = {
	'PyString' : 'O!',
	'int' : 'i',
	'PyObject *' : 'O',
	'float' : 'f',
	'double' : 'd'
		}

#
# What's the additional argument for the pythonType?
#
converter_lookup = {
	'PyString' : '&PyString_Type'
		}

#
# What's the manual (non argstring) conversion for the pythonType?
#
manual_conversion_lookup = {
	'double' : 'PyFloat_AsDouble(%s);',
	'float' : '(float)PyFloat_AsDouble(%s);',
	'PyObject *' : '%s;',
	'int' : '(int)PyInt_AsLong(%s);'
}

#
# How do I return the CType as a python representation?
#
return_as_python = {
	'PyObject *': "return %s;",
	'int': "return PyInt_FromLong((long)%s);",
	'double': "return PyFloat_FromDouble(%s);",
	'float': "return PyFloat_FromDouble(%s);",
	'void *': "return PyInt_FromLong((long)%s);"
		}

#
# How do I convert from the pythonType to the CType?
#
from_to = {'PyString': {'int *': "%s = (int *)PyString_AsString(%s);"}}

#
# Variables that are members of classes.  Also parameters for functions.
#
class TipItem:
	def __init__(this, args):
		this.name = args["name"]
		this.type = args["type"]
		this.values = \
			{"default": None, "onInit": 0, "pythonType": None,
			 "keyword": 0, "nokeyword": "*ERROR* supply nokeyword!",
			 "settable": 0, "gettable": 0, "definition": "default"}

		this.values.update(args)
		if this.values["pythonType"] == None:
			this.mainType = this.type
		else:
			this.mainType = this.values["pythonType"]
		this.set_text = []
	
	def set_set_text(this, text):
		this.set_text += text

	def is_primary_getter(this):
		return 0
	def is_primary_setter(this):
		return 0

	def is_optional(this):
		return (not this.values["default"] is None) and \
			this.values["onInit"] == "1"

	#
	# Generate the real type of this variable
	#
	def realtype(this):
		if this.type[0] == "I":
			return "PyObject *"
		return this.type

	#
	# Generate a declaration of a variable of this type
	#
	def declaration(this, include_file):
		s = CodeString(this.realtype() + " " + this.name)
		if include_file:
			s += ";\n"
			return s
		
		#
		# If this type doesn't need managing, we set the default
		# ("nokeyword") value here.  Otherwise we'll set a default
		# of NULL and test-assign after keyword parsing
		#
		if this.values["keyword"]:
			if this.needs_managing():
				s += " = NULL"
			else:
				s += " = %s" % (this.values["nokeyword"],)

		#
		# optional values have defaults, hence we should set that
		# default here
		#
		elif this.is_optional():
			s += " = %s" % (this.values["default"],)
		s += ";\n"
		
		if this.values["pythonType"] == None:
			return s
		
		s += "%s %s;\n" % (pt_lookup[this.values["pythonType"]],
			this.name + "_from")
		return s
	
	#
	#
	#
	def set_nokeyword_value(this):
		s = CodeString("if (%s == NULL)\n" % (this.name))
		s.indent()
		s += "%s = %s;\n" % (this.name, this.values["nokeyword"])
		return s
	
	#
	# Generate structure initialisation code from input
	#
	def variable_initialisation(this):
		s = CodeString()
		if this.type[0] == "I":
			s += "assert(IS_%s(%s));\n" % (this.type.upper(),
				this.name)
		s += "object->%s = %s;\n" % (this.name, this.name)
		return s
		
	#
	# Generate structure initialisation code from constant
	#
	def constant_initialisation(this):
		return CodeString("object->%s = %s;\n" % 
			(this.name, this.values["default"]))

	#
	# Generate a PyArg_ParseTuple format string fragment for this item
	#
	def parse_tuple_format(this):
		try:
			return pae_lookup[this.mainType]
		except:
			pass
		try:
			interface = all_interfaces[this.mainType]
			return "O"
		except:
			pass

		raise UnknownTypeException(this.mainType)
		
	def get_calling_arguments(this):
		return this.realtype() + " " + this.name

	#
	# Generate the inputs to the function PyArg_ParseTuple (or 
	# PyArg_ParseTupleAndKeywords).  if this is an object that needs
	# an object type provided, add that too.  If this is an object that
	# we are automatically converting into a C type, then we need to
	# put the output into name_from rather than name
	#
	def parse_tuple_inputs(this):
		s = CodeString()
		if this.parse_tuple_format() == "O!":
			s += converter_lookup[this.mainType] + ", "
		if this.values["pythonType"] is None:
			s += "&" + this.name
		else:
			s += "&" + this.name + "_from"
		return s
	
	#
	# Code for conversion from Python type to C type
	#
	def conversion_code(this):
		line = from_to[this.values["pythonType"]][this.type]
		s = CodeString(line % (this.name, this.name + "_from"))
		s += "\n"
		return s
	
	#
	# code for management of Python objects
	#
	def management_code(this, location):
		return CodeString("Py_XINCREF(%s);\n" % (location,))
	def unmanagement_code(this, location):
		return CodeString("Py_XDECREF(%s);\n" % (location,))
	#
	# Does this item need DECREF and INCREF management?
	#
	def needs_managing(this):
		return this.type == "PyObject *" or this.type[0] == "I"
	def generate_setget(this, className):
		s = this.generate_set(className) 
		s += this.generate_get(className)
		return s
	def generate_set(this, className):
		if not this.values["settable"] == "1":
			return CodeString("")
		#
		# Build C setter
		#
		s = CodeString("void C_%s_Set%s(PyObject *this, %s value) {\n" %
			(className, this.name, this.realtype()))
		s.indent()
		s += "%s *object = (%s *)this;\n\n" % (className, className)
		
		if this.needs_managing():
			s += "Py_XDECREF(object->%s);\n" % (this.name,)
			s += "Py_XINCREF(value);\n\n"
		for line in this.set_text:
			s += line + "\n"
		s += "object->%s = value;\n" % (this.name,)
		s.outdent()
		s += "}\n\n"

		#
		# Build python setter
		#
		s += "PyObject *%s_Set%s" % (className, this.name)
		s += "(PyObject *this, PyObject *args) {\n"
		s.indent()
		s += this.declaration(0)
		s += "\n"
		s += "if (!PyArg_ParseTuple(args, \"%s\", " % \
			(this.parse_tuple_format(),)
		s += this.parse_tuple_inputs()
		s += ")) {\n"
		s.indent()
		s += "return NULL;\n"
		s.outdent()
		s += "}\n\n"
		s += "C_%s_Set%s(this, %s);\n\n" % \
			(className, this.name, this.name)
		s += "Py_INCREF(Py_None);\n"
		s += "return Py_None;\n"
		s.outdent()
		s += "}\n"
		return s
	def generate_get(this, className):
		if not this.values["gettable"] == "1":
			return CodeString("")
		#
		# Build C getter
		#
		s = CodeString("%s C_%s_Get%s(PyObject *this) {\n" %
			(this.realtype(), className, this.name))
		s.indent()
		s += "%s *object = (%s *)this;\n\n" % (className, className)
		if this.needs_managing():
			s += "Py_XINCREF(object->%s);\n" % (this.name,)
		s += "return object->%s;\n" % (this.name,)
		s.outdent()
		s += "}\n\n"
		#
		# Build python getter
		#
		s += "PyObject *%s_Get%s" % (className, this.name)
		s += "(PyObject *this, PyObject *args) {\n"
		s.indent()
		s += this.declaration(1)
		s += "\n"
		s += "%s = C_%s_Get%s(this);\n\n" % \
			(this.name, className, this.name)

		if this.type[0] == "I":
			type = "PyObject *"
		else:
			type= this.type

		s += return_as_python[type] % (this.name,)
		s += "\n"
		s.outdent()
		s += "}\n"
		return s
	
	#
	# Generate records in the function definition struct for setter and 
	# getter for this item
	#
	def generate_setget_record(this, className):
		s = this.generate_set_record(className)
		s += this.generate_get_record(className)
		return s
	def generate_set_record(this, className):
		if not this.values["settable"] == "1":
			return CodeString()
		s = CodeString('{"set%s", (PyCFunction)%s_Set%s, METH_VARARGS, '
			% (this.name[0].upper() + this.name[1:], 
				className, this.name))
		s += '""},\n'
		return s
	def generate_get_record(this, className):
		if not this.values["gettable"] == "1":
			return CodeString()
		s = CodeString('{"get%s", (PyCFunction)%s_Get%s, METH_VARARGS, '
			% (this.name[0].upper() + this.name[1:], 
				className, this.name))
		s += '""},\n'
		return s
	def getCallArguments(this):
		return this.name

MANAGED=0
FIXED=1
USER=2
FIXED_RUNTIME=3

class TipItemArray:
	def __init__(this, args):
		this.name = args["name"]
		this.type = args["type"]
		this.values = \
			{"default": None, "onInit": 0, "pythonType": None,
			 "keyword": 0, "nokeyword": "*ERROR* supply nokeyword!",
			 "settable": 0, "gettable": 0, "size":"-1",
			 "arrayType": "*ERROR* specify array type", 
			 "primary": "0", "definition" : "default", 
			 "memory" : "static"}

		this.values.update(args)
		if this.values["pythonType"] == None:
			this.mainType = this.type
		else:
			this.mainType = this.values["pythonType"]
		size = int(this.values["size"])
		if this.type[0] == "I":
			this.type = "PyObject *"

		#
		# Possible modes:
		#	Managed array of unknown size (size=0)
		#		default=empty -> empty list (D)
		#	type=MANAGED
		#
		#	Unmanaged array of fixed size (size=n)
		#		onInit=elements -> elements are passed into
		#			initialiser
		#		onInit=0 -> elements are not var initialised (D)
		#			default=n -> elements are inited to n
		#	type=FIXED
		#	(or FIXED_RUNTIME if memory is allocated at runtime)
		#
		#	Unmanaged array of unknown size (size=0)
		#		default=empty -> empty list (D)
		#		user MUST specify:
		#		for settable: onSet onAlloc
		#		for gettable: onGet
		#	type=USER
		#	
		if this.values["arrayType"] == "managed":
			this.mode = MANAGED
			if this.values["onInit"] != 0:
				raise BadArraySpecificationException(
					this.name + "(don't specify onInit for"
					+ "managed arrays)")
		elif this.values["arrayType"] == "fixed":
			this.mode = FIXED
		elif this.values["arrayType"] == "user":
			this.mode = USER
		elif this.values["arrayType"] == "fixed_runtime":
			this.mode = FIXED_RUNTIME
		else:
			raise BadArraySpecificationException(
				this.name + " doesn't appear to be specified" +
				" correctly")
				
		this.get_text = []
		this.set_text = []
		this.alloc_text = []
	
	def is_optional(this):
		return 0
	
	def is_primary_getter(this):
		return this.values["primary"] == "1" and \
			this.values["gettable"] == "1"
	def is_primary_setter(this):
		return this.values["primary"] == "1" and \
			this.values["settable"] == "1"
	
	#
	# set additional code to be run when an element of this array is
	# assigned to
	#
	def set_set_text(this, lines):
		this.set_text += lines
	
	#
	# set additional code to be run on (re)allocation of this array
	#
	def set_alloc_text(this, lines):
		this.alloc_text += lines

	#
	# set additional code to be run when an element of this array
	# is retrieved
	#
	def set_get_text(this, lines):
		this.get_text += lines

	def declaration(this, interface_file):
		s = CodeString()
		if interface_file or not this.values["onInit"] == "elements":
			if this.mode == FIXED and \
				this.values["memory"] == "static":
				# i.e. I have a fixed, nonmanaged inline
				# array
				s += "%s %s[%s];\n" % (this.type, this.name,
					this.values["size"])
			else:
				# i.e. I have a not fixed array
				s += "%s *%s;\n" % (this.type, this.name)
				if not this.values["memory"] == "allocated":
					# i.e. it's not dynamic because of
					# a request for the memory to be
					# malloced
					s += "int %sSize;\n" % (this.name,)
					s += "int %sAlloc;\n" % (this.name,)
			return s
			
		size = int(this.values["size"])
		s += reduce(lambda a, b: a + ";\n" + b, map(lambda a: 
			"%s %s%d" % (this.type, this.name, a), range(size)))
		s += ";\n"
		
		return s
	
	def element_declaration(this):
		return '%s %s;\n' % (this.type, this.name)
	
	def get_calling_arguments(this):
		if this.values["onInit"] == "elements":
			size = int(this.values["size"])
			return reduce(lambda a, b: a + ", " + b, map(lambda a:
				"%s %s%d" % (this.type, this.name, a),
					range(size)))
		return this.type + " " + this.name
			
	#
	# Generate structure initialisation code from input
	#
	def variable_initialisation(this):
		size = int(this.values["size"])
		s = CodeString()
		if this.mode == FIXED_RUNTIME:
			s += "object->%s = malloc(%d * sizeof(%s));\n" % \
				(this.name, size, this.type)
		if this.values["onInit"] == "elements":
			s += reduce(lambda a, b: a + b,
				map(lambda a: "object->%s[%d] = %s%d;\n" %
					(this.name, a, this.name, a),
						range(size)))
		return s

	#
	# Generate structure initialisation code from constant
	#
	def constant_initialisation(this):
		if this.mode == FIXED_RUNTIME:
			size = int(this.values["size"])
			s = CodeString("object->%s = malloc(" % (this.name))
			s += "%d * sizeof(%s));\n" % (size, this.type)
			return s
		elif this.mode == MANAGED:
			s = CodeString("object->%s = NULL;\n" % (this.name))
			s += "object->%sSize = 0;\n" % (this.name)
			s += "object->%sAlloc = 0;\n" % (this.name)
			return s
		elif this.mode == USER:
			return CodeString("object->%s = NULL;\n" % (this.name))
		
	

	
	def needs_managing(this):
		return this.type == "PyObject *" or this.type[0] == "I"
	#
	# Generate a PyArg_ParseTuple format string fragment for this item
	#
	def parse_tuple_format(this):
		if this.values["onInit"] == "elements":
			size = int(this.values["size"])
			return pae_lookup[this.type] * size
		else:
			return pae_lookup[this.type + " *"]
			
	#
	# Generate the inputs to the function PyArg_ParseTuple (or 
	# PyArg_ParseTupleAndKeywords).  if this is an object that needs
	# an object type provided, add that too.  If this is an object that
	# we are automatically converting into a C type, then we need to
	# put the output into name_from rather than name
	#
	def parse_tuple_inputs(this):
		s = CodeString()
		if this.values["onInit"] == "elements":
			size = int(this.values["size"])
			s += reduce(lambda a, b: a + ", " + b, map(
				lambda a: "&%s%d" % (this.name, a), 
					range(size)))
			return s

	def getCallArguments(this):
		if this.values["onInit"] == "elements":
			size = int(this.values["size"])
			return reduce(lambda a, b: a + ", " + b, map(lambda a:
				"%s%d" % (this.name, a), range(size)))
		return this.name
	def generate_length(this, classname):
		if not this.values["primary"] == "1":
			return CodeString()
		s = CodeString("int %s_Length(PyObject *self) {\n" %
			(classname,))
		s.indent()
		if this.mode == MANAGED:
			s += "%s *object = (%s *)self;\n" % \
				(classname, classname)
			s += "return object->%sSize;\n" % (this.name,)
		elif not this.mode == USER:
			s += "return %d;\n" % (int(this.values["size"]),)
		s.outdent()
		s += "}\n"
		return s
		
	def generate_setget(this, classname):
		s = CodeString("")
		if this.values["settable"] == "1":
			s += this.generate_set(classname)
			if this.mode == MANAGED or this.mode == USER:
				s += this.generate_append(classname)
		if this.values["gettable"] == "1":
			s += this.generate_get(classname)
		s += this.generate_length(classname)
		return s
	
	def range_check(this, r_str="return NULL;\n"):
		if this.mode == USER:
			return CodeString()
		s = CodeString("if (index < 0 || index > ")
		if this.mode == FIXED or this.mode == FIXED_RUNTIME:
			s += str(int(this.values["size"]) - 1)
		else:
			s += "object->%sSize - 1" %(this.name,)
		s += ") {\n"	
		s.indent()
		s += 'PyErr_SetString(PyExc_IndexError, ' +\
			'"Array index out of range");\n'
		s += r_str
		s.outdent()
		s += "}\n\n"
		return s
	
	def generate_set(this, classname):
		s = CodeString("PyObject *%s_%s_AssItem(PyObject *self, "
			% (classname,this.name) + 
			"PyObject *args) {\n\n")
		s.indent()
		s += "int index;\n\n"
		s += this.element_declaration()
		s += 'if (!PyArg_ParseTuple(args, "i%s", &index, &%s)) {\n' % \
			(pae_lookup[this.type], this.name)
		s.indent()
		s += "return NULL;\n"
		s.outdent()
		s += "}\n\n"
		s += "%s *object = (%s *)self;\n\n" % (classname, classname)
		s += this.range_check()
		if this.needs_managing():
			s += "Py_XDECREF(object->%s[index]);\n" % (this.name,)
			s += "Py_XINCREF(%s);\n\n" % (this.name,)

		s += "object->%s[index] = %s;\n" % (this.name, this.name)
		for line in this.set_text:
			s += line + "\n"
		s += "\nPy_INCREF(Py_None);\n"
		s += "return Py_None;\n"
		s.outdent()
		s += "}\n\n"
	
		if not this.values["primary"] == "1":
			return s
			
		s += "int %s_AssItem(PyObject *self, " \
			% (classname,) + "int index, PyObject *input) {\n\n"
		s.indent()
		
		s += "%s *object = (%s *)self;\n\n" % (classname, classname)
		s += this.range_check("return 1;\n")
		s += "%s %s = %s\n" % (this.type, this.name, 
			(manual_conversion_lookup[this.type] % ("input",)))
		
		for line in this.set_text:
			s += line + "\n"
		if this.needs_managing():
			s += "Py_XDECREF(object->%s[index]);\n" % (this.name,)
			s += "Py_XINCREF(%s);\n\n" % (this.name,)
		
		s += "object->%s[index] = %s;\n" % (this.name, this.name)
		s += "\nreturn 0;\n"
		s.outdent()
		s += "}\n\n"
		return s

	def generate_append(this, classname):
		s = CodeString("PyObject *%s_Append(PyObject *self, " %
			(classname,))
		s += "PyObject *args) {\n"
		s.indent()
		s += this.element_declaration()

		s += 'if (!PyArg_ParseTuple(args, "%s", &%s)) {\n' % \
			(pae_lookup[this.type], this.name)
		s.indent()
		s += "return NULL;\n"
		s.outdent()
		s += "}\n\n"

		s += "if (!C_%s_Append(self, %s))\n\n" % (classname, this.name)
		s.indent()
		s += "return NULL;\n"
		s.outdent()
		s += "Py_INCREF(Py_None);\n"
		s += "return Py_None;\n"

		s2 = CodeString("int C_%s_Append(PyObject *self, %s %s);\n\n"  
			% (classname, this.type, this.name))
		s2.append_to("%s_common.h" % (classname,));
		
		s.outdent()
		s += "}\n\n"

		s += "int C_%s_Append(PyObject *self, %s %s) {\n" % \
			(classname, this.type, this.name)
		s.indent()

		s += "%s *object = (%s *)self;\n\n" % (classname, classname)
		
		if (this.values["type"][0] == "I"):
			s += "if (((TograMin *)%s)->interfaces == NULL) {\n" % \
				(this.name,)
			s.indent()
			s += 'PyErr_SetString(PyExc_TypeError, '
			s += '"Appended object is not a Togra object!");\n'
			s += "return 0;\n";
			s.outdent()
			s += "}\nif (!IS_%s(%s)) {\n" % (
				this.values["type"].upper(), this.name)
			s.indent()
			s += 'PyErr_SetString(PyExc_TypeError, '
			s += '"Appended object does not have %s' % (
				this.values["type"])
			s += ' interface!");\n'
			s += "return 0;\n";
			s.outdent()
			s += "}\n"

		if this.mode == MANAGED:
			s += "if (object->%sSize == object->%sAlloc) {\n" % \
				(this.name, this.name)
			s.indent()
			s += "if (object->%sAlloc == 0) {\n" % (this.name,)
			s.indent()
			s += "object->%sAlloc = 1;\n" % (this.name,)
			s.outdent()
			s += "} else {\n"
			s.indent()
			s += "object->%sAlloc *= 2;\n" % (this.name,)
			s.outdent()
			s += "}\n"
			s += "object->%s = realloc(object->%s, " % \
				(this.name, this.name)
			s += "object->%sAlloc * sizeof(%s));\n" % \
				(this.name,this.type)
		s.outdent()
		for line in this.alloc_text:
			s += line + "\n"
		s += "}\n\n"
		s += "int index = object->%sSize++;\n\n" % (this.name,)
		for line in this.set_text:
			s += line + "\n"
		if this.needs_managing():
			s += "Py_XINCREF(%s);\n\n" % (this.name,)
		s += "object->%s[index] = %s;\n" % \
			(this.name, this.name)
		s += "return 1;\n"
		s.outdent()
		s += "}\n\n"
		return s
	
	def generate_get(this, classname):
		
		s = CodeString("PyObject *%s_%s_Item(PyObject *self, " %
			(classname, this.name) + "PyObject *args) {\n\n")
		s.indent()
		s += "int index;\n\n"
		s += 'if (!PyArg_ParseTuple(args, "i", &index)) {\n'
		s.indent()
		s += 'return NULL;\n'
		s.outdent()
		s += "}\n\n"
	
		s += "%s *object = (%s *)self;\n\n" % (classname, classname)
		s += this.range_check()
		
		if this.needs_managing():
			s += "Py_XINCREF(object->%s[index]);\n" % (this.name,)

		for line in this.get_text:
			s += line + "\n"

		s += return_as_python[this.type] % \
			("object->%s[index]" % (this.name,))
		s += "\n"
		s.outdent()
		s += "}\n\n"
	
		if not this.values["primary"] == "1":
			return s
		
		s += "PyObject *%s_Item(PyObject *self, " % \
			(classname,) + "int index) {\n\n"
		s.indent()
		s += "%s *object = (%s *)self;\n\n" % (classname, classname)
		s += this.range_check()
		
		if this.needs_managing():
			s += "Py_XINCREF(object->%s[index]);\n" % (this.name,)

		s += return_as_python[this.type] % \
			("object->%s[index]" % (this.name,))
		s.outdent()
		s += "\n}\n\n"
		return s
	def generate_setget_record(this, classname):
                s = this.generate_set_record(classname)
                s += this.generate_get_record(classname)
                return s
	def generate_set_record(this, classname):
		if not this.values["settable"] == "1":
			return CodeString()
		s = CodeString('{"set%sItem", (PyCFunction)%s_%s_AssItem, ' % \
			(this.name[0].upper() + this.name[1:], classname,
				this.name))
		s += 'METH_VARARGS, ""},\n'
		if this.mode == MANAGED or this.mode == USER:
			s += '{"append", (PyCFunction)%s_Append, ' % \
				(classname,)
			s += 'METH_VARARGS, ""},\n'
		return s
	def generate_get_record(this, classname):
		if not this.values["gettable"] == "1":
			return CodeString()
		s = CodeString('{"get%sItem", (PyCFunction)%s_%s_Item, ' % \
			(this.name[0].upper() + this.name[1:], classname,
				this.name))
		s += 'METH_VARARGS, ""},\n'
		return s


class Tip(DictParser):
	def __init__(this, file):
		DictParser.__init__(this, file)
		this.classes = []
		this.globals = []

		this.add_tag("root", DictEntry(None, None, this.finished))

		this.add_tag("interfaceDefinitions",
			DictEntry(None, None, this.done_interface_defs),
			["root"])
		this.add_tag("interfaceDefinition", 
			DictEntry(this.on_new_ifdef, None, this.done_ifdef),
			["root", "interfaceDefinitions"])
		this.add_tag("function",
			DictEntry(this.on_new_idfunc, None, this.done_idfunc),
			["root", "interfaceDefinitions", "interfaceDefinition"])
		this.add_tag("parameter",
			DictEntry(this.add_idfunc_parameter, None, None),
			["root", "interfaceDefinitions", "interfaceDefinition",
			"function"])
		this.add_tag("returnNone",
			DictEntry(this.idfunc_noreturn, None, None),
			["root", "interfaceDefinitions", "interfaceDefinition", 
			"function"])
		this.add_tag("return",
			DictEntry(this.idfunc_return, None, None),
			["root", "interfaceDefinitions", "interfaceDefinition", 
			"function"])
			
		this.add_tag("tograClass", DictEntry(this.on_new_class,
			None, this.done_class), ["root"])
		this.add_tag("interfaces", DictEntry(None, None, 
			this.done_interfaces), ["root", "tograClass"])
		this.add_tag("interface", DictEntry(this.on_new_interface,
			None, None), ["root", "tograClass", "interfaces"])
		this.add_tag("pythonFunctions", DictEntry(None, None, None),
			["root", "tograClass"])
		this.add_tag("extern", 
			DictEntry(this.on_new_extern, None, None),
			["root", "tograClass", "pythonFunctions"])
		this.add_tag("defineGroup",
			DictEntry(this.on_new_define_group, None, 
			this.done_define_group),
			["root", "tograClass", "pythonFunctions"])
		this.add_tag("define", DictEntry(this.add_define, None, None),
			["root", "tograClass", "pythonFunctions", 
				"defineGroup"])
		this.add_tag("function", DictEntry(this.on_new_function,
			this.function_text, this.done_function), 
			["root", "tograClass", "pythonFunctions"])
		this.add_tag("parameter", DictEntry(this.on_new_parameter,
			None, None), ["root", "tograClass", "pythonFunctions",
			"function"])
		this.add_tag("returnNone", DictEntry(this.noreturn, None,
			None), ["root", "tograClass", "pythonFunctions", 
			"function"])
		this.add_tag("return", DictEntry(this.somereturn, None,
			None), ["root", "tograClass", "pythonFunctions", 
			"function"])
		this.add_tag("storedItems", DictEntry(None, None, None),
			["root", "tograClass"])
		this.add_tag("item", DictEntry(this.on_new_item, None, None),
			["root", "tograClass", "storedItems"])
		this.add_tag("onSet", DictEntry(None, this.set_text, None),
			["root", "tograClass", "storedItems", "item"])
		this.add_tag("itemArray", 
			DictEntry(this.on_new_item_array, None, None),
			["root", "tograClass", "storedItems"])
		this.add_tag("onSet", DictEntry(None, this.set_text, None),
			["root", "tograClass", "storedItems", "itemArray"])
		this.add_tag("onAlloc", DictEntry(None, this.alloc_text, None),
			["root", "tograClass", "storedItems", "itemArray"])
		this.add_tag("onGet", DictEntry(None, this.get_text, None),
			["root", "tograClass", "storedItems", "itemArray"])
		this.add_tag("onInitialise", DictEntry(None, this.init_text,
			None), ["root", "tograClass", "storedItems"])
		this.add_tag("extraInput", DictEntry(this.extra_init_input,
			None, None), ["root", "tograClass", "storedItems",
			"onInitialise"])
		this.add_tag("onFree", DictEntry(None, this.free_text,
			None), ["root", "tograClass", "storedItems"])
		this.add_tag("globalFunctions", DictEntry(None, None, None),
			["root"])
		this.add_tag("function", DictEntry(this.new_global, None, None),
			["root", "globalFunctions"])

		this.interfaces = {}
		this.all_data = ""

	#
	# New interface definition.  Set up an entry in the interfaces
	# dictionary.
	#
	def on_new_ifdef(this, name, props):
		this.curInterface = TipInterface(props)
		this.strip_data = None
	
	def on_new_idfunc(this, name, props):
		this.curFunction = TipFunction(props)
		this.strip_data = None
	
	def add_idfunc_parameter(this, name, props):
		parm = TipItem(props)
		this.curFunction.add_parameter(parm)

	def idfunc_noreturn(this, name, props):
		this.curFunction.returnType = None
	
	def idfunc_return(this, name, props):
		this.curFunction.returnType = props["type"]
	
	def done_idfunc(this, name):
		this.curInterface.add_function(this.curFunction)

	def done_ifdef(this, name):
		this.interfaces[this.curInterface.name] = this.curInterface

	#
	#  Compile list of interfaces and build interface struct
	#
	def done_interface_defs(this, name):
		# create individual interface structures
		global all_interfaces
		all_interfaces = this.interfaces
		s = CodeString('#include "TograIncludes.h"\n\n')
		for interface in this.interfaces.values():
			s += interface.struct_definition()

		# create interface list
		s += "typedef struct {\n"
		s.indent()
		for interface in this.interfaces.values():
			s += interface.name + " *" + interface.name + ";\n"
		s.outdent()
		s += "} IC;\n\n"

		# create TograMin object
		s += "typedef struct {\n\tPyObject_HEAD\n\tIC *interfaces;"
		s += "\n} TograMin;\n\n"

		for interface in this.interfaces.values():
			s += interface.accessor_defines()
		
		s.write_to("interfaces.h")
		
		s = CodeString()
		s.write_to("definitions.h")

		for line in copyright_notice:
			s += line

		s += '\n#include "TograIncludes.h"\n\n'
		
		s.write_to("Togra.c")
	
	def new_global(this, name, props):
		this.globals.append(props)
		
		s = CodeString("\nPyObject")
		s += " *%s(PyObject *self, PyObject *args);\n" % \
			(props["name"],)
		s.append_to("Togra.c")
	
	def on_new_extern(this, name, props):
		this.current_class.add_extern(
			TipExtern(props["name"], props["type"]))
		this.strip_data = None

	def on_new_define_group(this, name, props):
		this.current_define_group = TipDefineGroup()
		this.strip_data = None
		
	def add_define(this, name, props):
		this.current_define_group.add_item(props["name"])
		
	def done_define_group(this, name):
		this.current_class.add_define_group(this.current_define_group)
		this.current_define_group = None
	#
	# New Class Definition
	#
	def on_new_class(this, name, props):
		this.current_class = TipClass(props)
		this.strip_data = None
	
	def done_class(this, name):
		this.current_class.generate_class_code(this.interfaces)
		this.classes.append(this.current_class)

	def done_interfaces(this, name):
		pass
	
	def on_new_interface(this, name, props):
		this.current_class.add_interface(this.interfaces[props["name"]])
		this.strip_data = None

	def on_new_function(this, name, props):
		this.current_function = TipFunction(props)
		this.strip_data = None

	#
	# should try to clean the text up a bit.  The initial indentation will 
	# be specified by the code that TipParser dumps, so I'll remove all
	# of the initial indentation and try to do the same to each other line
	#
	# I also convert a couple of special symbols to C equivalents.  This
	# gets around element painfullness in XML.  The symbols are:
	#	~REF	->	&
	#	~LT	->	<
	def clean_text(this, data):
		lines = data.split('\n')
		clean_lines = []
		for line in lines:
			if line.strip() == "":
				continue
			
			if this.strip_data == None:
				this.strip_data = ""
				while line[0].isspace():
					this.strip_data += line[0]
					line = line[1:]
			else:
				try:
					cli = line.index(this.strip_data)
					if (cli == 0):
						l = len(this.strip_data)
						line = line[l:]
				except:
					pass
			while 1:
				try:
					i = line.index("~REF")
				except:
					break
				line = line[:i] + "&" + line[i+4:]
			while 1:
				try:
					i = line.index("~LT")
				except:
					break
				line = line[:i] + "<" + line[i+3:]
			clean_lines.append(line)
		return clean_lines
	
	def function_text(this, data):
		this.all_data += data
		if this.all_data[-1] == "\n":
			clean_lines = this.clean_text(this.all_data)
			this.all_data = ""
			this.current_function.set_text(clean_lines)
				
	
	def done_function(this, name):
		this.current_class.add_function(this.current_function)
	
	def on_new_parameter(this, name, props):
		parm = TipItem(props)
		this.current_function.add_parameter(parm)

	def noreturn(this, name, props):
		this.current_function.returnType = None
	
	def somereturn(this, name, props):
		this.current_function.returnType = props["type"]
	
	def on_new_item(this, name, props):
		item = TipItem(props)
		this.current_item = item
		this.current_class.add_item(item)

	def on_new_item_array(this, name, props):
		itemArray = TipItemArray(props)
		this.current_item = itemArray
		this.current_class.add_item(itemArray)

	def set_text(this, data):
		this.all_data += data
		if this.all_data[-1] == "\n":
			clean_lines = this.clean_text(this.all_data)
			this.all_data = ""
			this.current_item.set_set_text(clean_lines)
	
	def alloc_text(this, data):
		this.all_data += data
		if this.all_data[-1] == "\n":
			clean_lines = this.clean_text(this.all_data)
			this.all_data = ""
			this.current_item.set_alloc_text(clean_lines)

	def get_text(this, data):
		this.all_data += data
		if this.all_data[-1] == "\n":
			clean_lines = this.clean_text(this.all_data)
			this.all_data = ""
			this.current_item.set_get_text(clean_lines)

	def init_text(this, data):
		this.all_data += data
		if this.all_data[-1] == "\n":
			clean_lines = this.clean_text(this.all_data)
			this.all_data = ""
			this.current_class.set_init_text(clean_lines)
	
	def free_text(this, data):
		this.all_data += data
		if this.all_data[-1] == "\n":
			clean_lines = this.clean_text(this.all_data)
			this.all_data = ""
			this.current_class.set_free_text(clean_lines)

	def extra_init_input(this, name, props):
		item = TipItem(props)
		this.current_class.add_init_input(item)
	
	def finished(this, name):
		s = CodeString("\nPyMethodDef Togra_methods[] = {\n")
		s.indent()

		for cclass in this.classes:
			s += cclass.generate_methodDef()

		for gglobal in this.globals:
			s += '{"%s", %s, METH_VARARGS, ""},\n' % \
				(gglobal["pythonName"], gglobal["name"])

		s += '{NULL, NULL, 0, NULL}\n'
		s.outdent()
		s += "};\n\n"

		s += 'DL_EXPORT(void) initTogra(void) {\n'
		s.indent()
		s += 'InitTime();\n'
		s += 'Py_InitModule("Togra", Togra_methods);\n'
		s.outdent()
		s += '}\n\n'

		s.append_to("Togra.c")

tipfile = sys.argv[1]
f = open(tipfile, 'r')
dp = Tip(f)
try:
	dp.go()
except TipException, inst:
	print inst.string
	sys.exit(1)
f.close()
