/*
 * Copyright 2003, Shane Stephens, Andre Pang.  
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

#include "TograIncludes.h"

#include "StaticRender_Privates.h"

staticforward PyTypeObject PyContainerType;

static void C_PyContainer_Render(PyObject *self, int mode, int picking);
static int C_PyContainer_Compile(PyObject *self);

static IRender PyContainer_IRender = {C_PyContainer_Render, 
                C_PyContainer_Compile};
static IC PyContainer_Interfaces = {&PyContainer_IRender, NULL};

int PyContainer_Check(PyObject *v) {
 
        return (v->ob_type == &PyContainerType);
}

PyObject *C_PyContainer_New(void) {
	
        PyContainer *c 
                = (PyContainer *)PyObject_New(PyContainer, &PyContainerType);

        D_ALLOC(c, "Container");

        c->interfaces = &PyContainer_Interfaces;
	c->alloced = c->length = 0;
	c->transform = C_PyTransformContainer_New();
	c->children = NULL;
        c->status = NULL;
        c->is_static = 1;
        c->display_list = 0;

        c->material = NULL;

        Py_INCREF(c->transform);
        
	return (PyObject *)c;
}

PyObject *C_PyContainer_OfLength(int length) {

	PyContainer *c = (PyContainer *)C_PyContainer_New();
	c->alloced = length;
	c->children = malloc(sizeof(PyObject *) * length);

	return (PyObject *)c;
	
}

PyObject *PyContainer_New(PyObject *self, PyObject *args) {
	
        int length;
        PyObject *o;

        
        if (PyTuple_Size(args) == 0) {
		return (PyObject *)C_PyContainer_New();
	}
	
        if (!PyArg_ParseTuple(args, "i", &length)) {
		return NULL;
	}

        if (length == 0)
                return (PyObject *)C_PyContainer_New();

        o = (PyObject *)C_PyContainer_OfLength(length);
        return o;
	
}

static void C_PyContainer_Free(PyObject *self) {
	
 	PyContainer *c = (PyContainer *)self;
	int i;

        D_FREE(self, "Container");
        
	if (c->alloced > 0) {
		
		for (i = 0; i < c->length; i++) {
			Py_DECREF(c->children[i]);
		}
		
		free(c->children);
	}

        if (!(c->status == NULL))
                free(c->status);
        
	PyObject_Del((PyObject *)c);
	
}

static void PyContainer_Free(PyObject *self) {
	
	C_PyContainer_Free(self);
	
}

void C_PyContainer_Append(PyObject *self, PyObject *obj) {

	PyContainer *c = (PyContainer *)self;
	
	Py_INCREF(obj);

        
	if (c->length == c->alloced) {
		if (c->alloced == 0) {
			c->alloced++;
		}
		else {
			c->alloced *= 2;
		}
		
                c->children = realloc(c->children, 
                                c->alloced * sizeof(PyObject *));
	}
	
        c->children[c->length++] = obj;

	
}

static PyObject *PyContainer_Append(PyObject *self, PyObject *args) {
	
	PyObject *obj;

	if (PyTuple_Size(args) != 1) {
		return NULL;
	}
	
	obj = PyTuple_GetItem(args, 0);
        
	if (obj == NULL) {
		return NULL;
	}
	
	C_PyContainer_Append(self, obj);
	
	Py_INCREF(Py_None);
	return Py_None;
	
}

static void C_PyContainer_SetMaterial(PyObject *self, PyObject *material) {
        
        PyContainer *c = (PyContainer *)self;

        Py_INCREF(material);
        if (c->material != NULL)
        {
                Py_DECREF(c->material);
        }

        c->material = material;
        
}
                

static PyObject *PyContainer_SetMaterial(PyObject *self, PyObject *args) {

        PyObject *m;
        
        if (!PyArg_ParseTuple(args, "O", &m))
        {
                return NULL;
        }

        C_PyContainer_SetMaterial(self, m);

        Py_INCREF(Py_None);
        return Py_None;
                
        
}

static PyMethodDef PyContainer_methods[] = {
        {"append", (PyCFunction)PyContainer_Append, METH_VARARGS,
                "Appends an object to the container"},
        {"setMaterial", (PyCFunction)PyContainer_SetMaterial, METH_VARARGS,
                "Set the default material for this container and descendants"}, 
        {NULL, NULL, 0, NULL}
};

static PyObject *PyContainer_GetAttr(PyObject *self, char *name) {

        if (strcmp(name, "transformContainer") == 0)
        {
                Py_INCREF(((PyContainer *)self)->transform);
                return ((PyContainer *)self)->transform;
        }
        
        return Py_FindMethod(PyContainer_methods, self, name);

}

static PyObject *PyContainer_Item(PyObject *self, int index) {
	
	PyContainer *c = (PyContainer *)self;
	
	if (index < c->length && index >= 0) {
		Py_INCREF(c->children[index]);
		return c->children[index];
	}

	
	return NULL;
	
}

static int PyContainer_SetAttr(PyObject *obj, char *name, PyObject *v) {
	
	PyContainer *r = (PyContainer *)obj;
	
	if (strcmp(name, "renderPath") == 0) {
                r->renderPath = v;
                Py_INCREF(v);
        } else if (strcmp(name, "transformContainer") == 0) {
		Py_DECREF(r->transform);
		r->transform = v;
		Py_INCREF(r->transform);
	} else {
                return -1;
        }
	
	return 0;
	
}

static inline void C_PyContainer_do_render(PyContainer *c, int time, 
			int picking) {
        
        int i;

	for (i = 0; i < c->length; i++) {
		RENDER(c->children[i], time, picking);
	}
}

static int C_PyContainer_Compile(PyObject *self) {

        PyContainer *c = (PyContainer *)self;
        int i;
	TPRINTF("Container %p compiling\n", self);
        for (i = 0; i < c->length; i++) {
                if (COMPILE(c->children[i]) == 0)
                        c->is_static = 0;
        }
        
#if 0
        if (c->is_static && c->display_list == 0)
        {
                c->display_list = glGenLists(1);
                glNewList(c->display_list, GL_COMPILE);
                C_PyContainer_do_render(c, 0); //time);
                glEndList();
        }
#endif
        return c->is_static;
        
}
        
static void C_PyContainer_Render(PyObject *self, int time, int picking) {
	
        PyContainer *c = (PyContainer *)self;

	glLoadName((int)self);
	
	glPushMatrix();

        if (c->transform != NULL) {
                TPRINTF("Container %p calling GENERATE\n", self);
		GENERATE(c->transform, c->status, time, 1, 0, 1);
	}

        if ((c->material != NULL) && (!picking))
        {
		TPRINTF("Container %p setting material %p\n", c, c->material);
                SET_MATERIAL(c->material, time);
        }
        
        if (0) { //c->is_static) {
                if (c->display_list == 0) {
                        C_PyContainer_Compile(self);
                }

                glCallList(c->display_list);

        } else {
                C_PyContainer_do_render(c, time, picking);
        }
        
        if ((c->material != NULL) && (!picking))
        {
		TPRINTF("Container %p unsetting material %p\n", c, c->material);
                UNSET_MATERIAL(c->material, time);
        } 
        
	glPopMatrix();

	glLoadName(0);
}
		
static PyObject *PyContainer_Call(PyObject *self, PyObject *args, PyObject *dict) {
	
	C_PyContainer_Render(self, 0, 0);
	
	Py_INCREF(Py_None);
	return Py_None;
}

static PySequenceMethods PyContainer_sequenceMethods = {
        0, /* length */
        0, /* concat */
        0, /* repeat */
        PyContainer_Item, /* item */
        0, /* slice */
        0, /* ass_item */
        0, /* ass_slice */
        0, /* contains */
        0, /* inplace_concat */
        0 /* inplace_repeat */
};

static PyTypeObject PyContainerType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PyContainer",
        sizeof(PyContainer),
        0,
        PyContainer_Free,
        0, /* print */
        PyContainer_GetAttr, /* getattr */
        PyContainer_SetAttr, /* setattr */
        0, /* compare */
        0, /* repr */
        0, /* as_number */
        &PyContainer_sequenceMethods, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        &PyContainer_Call, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};


staticforward PyTypeObject PySphereType;

PyObject *C_PySphere_New(void) {
	
        PyContainer *c 
                = (PyContainer *)PyObject_New(PySphere, &PySphereType);
	c->interfaces = &PyContainer_Interfaces;
	c->alloced = c->length = 0;
	c->transform = C_PyTransformContainer_New();
	c->children = NULL;
        c->renderPath = NULL;
        c->status = NULL;
        c->is_static = 1;
        c->display_list = 0;

        c->material = NULL;

        Py_INCREF(c->transform);
        
	return (PyObject *)c;
}

void C_PyFloatList_Append(PyObject *, float);

static void C_PySphere_SetMatCoords(PyObject *self) {

        PySphere *sp = (PySphere *)self;

        PyObject *top = C_PyFloatList_New();
        PyObject *bottom = C_PyFloatList_New();

        float slice_height = 2.0 / sp->slices;
        float cur_slice = 1 - slice_height;
        float cur_r = sqrt(1 - cur_slice * cur_slice);
        float x, x_last, old_r;
        int i, j;

        C_PyFloatList_Append(top, 0.5);
        C_PyFloatList_Append(top, 0.0);
        C_PyFloatList_Append(bottom, 0.5);
        C_PyFloatList_Append(bottom, 1.0);

        x = -cur_r/2.0;
        x_last = 0.5 + x;
        old_r = cur_r;

        
        for (i = 0; i < sp->segs + 1; i++) {
                C_PyFloatList_Append(top, 0.5 - x - (cur_r / (sp->segs + 1)));
                C_PyFloatList_Append(top, 1.0/sp->slices);
                C_PyFloatList_Append(bottom, 0.5 + x);
                C_PyFloatList_Append(bottom, 1 - 1.0/sp->slices);
                x += (cur_r / (sp->segs + 1));
        }

        
        C_PyStaticRender_SetTextureCoords(sp->children[0], top);
        C_PyStaticRender_SetTextureCoords(sp->children[sp->slices - 1], bottom);
        
        for (i = 0; i < sp->slices - 2; i++) {
                PyObject *temp = C_PyFloatList_New();
                
                cur_slice -= slice_height;

                cur_r = sqrt(1 - cur_slice * cur_slice);
                x = 0.5 - cur_r/2.0;

                for (j = 0; j < sp->segs + 1; j++) {
                        C_PyFloatList_Append(temp, x);
                        C_PyFloatList_Append(temp, (i + 2.0)/sp->slices);
                        C_PyFloatList_Append(temp, x_last);
                        C_PyFloatList_Append(temp, (i + 1.0)/sp->slices);
                        x += (cur_r / (sp->segs + 1));
                        x_last += (old_r / (sp->segs + 1));
                }

                C_PyStaticRender_SetTextureCoords(sp->children[i + 1], temp);
                x_last = 0.5 - cur_r/2.0;
                old_r = cur_r;
        }

}

                
static PyObject *PySphere_SetMatCoords(PyObject *self, PyObject *args) {

        C_PySphere_SetMatCoords(self);

        Py_INCREF(Py_None);
        return Py_None;
        
}

static PyMethodDef PySphere_methods[] = {
        {"setMaterialCoords", (PyCFunction)PySphere_SetMatCoords, METH_VARARGS,
                "Sets the material coordinates of the sphere"},
        {NULL, NULL, 0, NULL}
};

static PyObject *PySphere_GetAttr(PyObject *self, char *name) {

        PyObject *r = Py_FindMethod(PySphere_methods, self, name);

        if (r == NULL) {
                PyErr_Clear();
                return PyContainer_GetAttr(self, name);
        }

        return r;

}

static PyTypeObject PySphereType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PySphere",
        sizeof(PySphere),
        0,
        PyContainer_Free,
        0, /* print */
        PySphere_GetAttr, /* getattr */
        PyContainer_SetAttr, /* setattr */
        0, /* compare */
        0, /* repr */
        0, /* as_number */
        &PyContainer_sequenceMethods, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        &PyContainer_Call, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};
