/*
 * Copyright 2003-5, Shane Stephens.
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

#include "TograIncludes.h"

TOGRA_DEFINE_GRAPH(Graph)

void *C_PyGraph_New(void) {
	
	PyGraph *g = PyObject_New(PyGraph, &PyGraphType);
	D_ALLOC(g, "Graph");

	g->interfaces = &PyGraph_Interfaces;
	
	g->num_nodes = 0;
	g->num_connections = 0;
	g->num_groups = 0;
	g->nodes = NULL;
	g->connections = NULL;
	g->groups = NULL;
	g->callback = NULL;

	return g;
}

PyObject *PyGraph_New(PyObject *self, PyObject *args) {

	return C_PyGraph_New();

}

void C_PyGraph_Dump(PyGraph *g) {
	
	int i, j;
	for (i = 0; i < g->num_groups; i++)
	{
		GraphGroup *gr = g->groups[i];
		printf("Group %d\n", i);
		
		for (j = 0; j < gr->num_nodes; j++)
		{
			GraphNode *n = gr->nodes[j];
			printf("%p", n->object);

			if (n->group != gr)
				printf(" BAD GROUP!");
			printf("\n");
		}
		for (j = 0; j < gr->num_connections; j++)
		{
			GraphConnection *c = gr->connections[j];
			printf("%p-<%2.2f,%2.2f>-%p", c->from, c->length, 
					c->strength, c->to);
			if (c->group != gr)
				printf(" BAD GROUP!");
			printf("\n");
		}
	}
}
		

void C_PyGraph_AddGroupConnection(PyGraph *graph, GraphGroup *group,
		GraphConnection *c) {

	group->num_connections += 1;
	group->connections = realloc(group->connections,
			group->num_connections * sizeof(GraphConnection *));
	group->connections[group->num_connections - 1] = c;
	
}

void C_PyGraph_AddNode(PyGraph *graph, PyObject *node) {

	GraphGroup *g;
	GraphNode *n;
	
	graph->num_nodes += 1;
	graph->nodes = realloc(graph->nodes, 
			graph->num_nodes * sizeof(GraphNode *));

	n = malloc(sizeof(GraphNode));
	
	graph->nodes[graph->num_nodes - 1] = n;
	n->effected = 1;
	n->object = node;

	graph->num_groups += 1;
	graph->groups = realloc(graph->groups,
			graph->num_groups * sizeof(GraphGroup *));

	g = malloc(sizeof(GraphGroup));
		
	graph->groups[graph->num_groups - 1] = g;
	g->num_nodes = 1;
	g->nodes = malloc(sizeof(GraphNode *));
	g->nodes[0] = n;
	g->num_connections = 0;
	g->connections = NULL;

	n->group = g;

	if (graph->callback != NULL)
	{
		graph->callback(graph->data, GRAPH_NODE_ADDED, (void *)n);
	}
	
	Py_INCREF(node);

	//C_PyGraph_Dump(graph);

}

PyObject *PyGraph_AddNode(PyObject *self, PyObject *args) {

	PyObject *node;
	
	if (!PyArg_ParseTuple(args, "O", &node)) {
		return NULL;
	}

	C_PyGraph_AddNode((PyGraph *)self, node);

	Py_INCREF(Py_None);
	return Py_None;
	
}

void C_PyGraph_DeleteGroup(PyGraph *graph, GraphGroup *a)
{
	int i;

	for (i = 0; i < graph->num_groups; i++) {
		if (graph->groups[i] == a) {
			break;
		}
	}

	if (graph->num_groups > 1)
	{
		memcpy(graph->groups + i, graph->groups + i + 1, 
			(graph->num_groups - i - 1) * sizeof(GraphGroup *));
		graph->num_groups -= 1;
		graph->groups = realloc(graph->groups, 
			graph->num_groups * sizeof(GraphGroup *));
	}
	else {
		graph->num_groups = 0;
		free(graph->groups);
		graph->groups = NULL;
	}
}
				

void C_PyGraph_MergeGroups(PyGraph *graph, GraphGroup *a, GraphGroup *b)
{

	int i;

	for (i = 0; i < b->num_nodes; i++)
	{
		b->nodes[i]->group = a;
	}

	for (i = 0; i < b->num_connections; i++)
	{
		b->connections[i]->group = a;
	}
	
	/*
	 * There should always be nodes in both group a and group b, 
	 * otherwise they would already have been deleted
	 */
	a->nodes = realloc(a->nodes, sizeof(GraphNode *) * 
			(a->num_nodes + b->num_nodes));
	memcpy(a->nodes + a->num_nodes, b->nodes, 
			b->num_nodes * sizeof(GraphNode *));
	a->num_nodes += b->num_nodes;
	
	/*
	 * On the other hand, there may not be connections (singleton group)
	 */
	if (b->num_connections > 0) {
		if (a->connections == NULL) {
			a->connections = b->connections;
			a->num_connections = b->num_connections;
		} else {
			a->connections = realloc(a->connections, 
				sizeof(GraphConnection *) * 
				(a->num_connections + b->num_connections));
			memcpy(a->connections + a->num_connections,
				b->connections,
				b->num_connections * sizeof(GraphConnection *));
			a->num_connections += b->num_connections;
		}
	}

	C_PyGraph_DeleteGroup(graph, b);

}
	
int C_PyGraph_AddConnection(PyGraph *graph, PyObject *from, PyObject *to, 
		double length, double strength) {

	int i;
	GraphNode *from_index = NULL, *to_index = NULL;
	GraphConnection *c;
	
	for (i = 0; i < graph->num_nodes; i++)
	{
		if (graph->nodes[i]->object == from) {
			from_index = graph->nodes[i];
		} else if (graph->nodes[i]->object == to) {
			to_index = graph->nodes[i];
		}
	}
	
	if (from_index == NULL || to_index == NULL)
		return -1;
	
	graph->num_connections += 1;
	graph->connections = realloc(graph->connections,
			graph->num_connections * sizeof(GraphConnection *));

	c = malloc(sizeof(GraphConnection));
	
	graph->connections[graph->num_connections - 1] = c;
	c->from = from;
	c->to = to;
	c->length = length;
	c->strength = strength;

	if (from_index->group != to_index->group) {
		C_PyGraph_MergeGroups(graph, from_index->group, 
				to_index->group);
	}

	to_index->group = from_index->group;
	
	c->group = from_index->group;

	C_PyGraph_AddGroupConnection(graph, from_index->group, c);
	
	Py_INCREF(from);
	Py_INCREF(to);

	//C_PyGraph_Dump(graph);
	
	return 0;
	
}

PyObject *PyGraph_AddConnection(PyObject *self, PyObject *args) {

	PyObject *from, *to;
	double length, strength;
	
	if (!PyArg_ParseTuple(args, "OOdd", &from, &to, &length, &strength)) {
		return NULL;
	}

	if (C_PyGraph_AddConnection((PyGraph *)self, from, to, length, 
				strength) == -1) {
		return NULL;
	}

	Py_INCREF(Py_None);
	return Py_None;
	
}

int C_PyGraph_GetNodes(PyObject *self, GraphNode ***nodes) {

	PyGraph *graph = (PyGraph *)self;
	
	*nodes = graph->nodes;
	return graph->num_nodes;

}

int C_PyGraph_GetConnections(PyObject *self, GraphConnection ***connections) {

	PyGraph *graph = (PyGraph *)self;
	
	*connections = graph->connections;
	return graph->num_connections;

}

void C_PyGraph_RegisterCallback(PyObject *self, PyObject *data, 
		GraphCallbackEvent callback) {

	PyGraph *graph = (PyGraph *)self;

	graph->callback = callback;
	graph->data = data;
	Py_INCREF(data);

}


int C_PyGraph_GetNumGroups(PyObject *self) {

	PyGraph *graph = (PyGraph *)self;
	
	return graph->num_groups;

}

PyObject *PyGraph_GetNumGroups(PyObject *self, PyObject *args) {

	return PyInt_FromLong((long)C_PyGraph_GetNumGroups(self));

}

int C_PyGraph_GetGroupNodes(PyObject *self, int idx, GraphNode ***n) {

	PyGraph *graph = (PyGraph *)self;
	
	if (idx >= graph->num_groups || idx < 0) {
		*n = NULL;
		return -1;
	}

	*n = graph->groups[idx]->nodes;
	return graph->groups[idx]->num_nodes;

}

PyObject *PyGraph_GetGroupNodes(PyObject *self, PyObject *args) {

	int idx, i, num_nodes;
	GraphNode **nodes;
	PyObject *list;
	
	if (!PyArg_ParseTuple(args, "i", &idx)) {
		return NULL;
	}

	num_nodes = C_PyGraph_GetGroupNodes(self, idx, &nodes);
	
	if (nodes == NULL) {
		return NULL;
	}

	list = PyList_New(0);
	for (i = 0; i < num_nodes; i++)
	{
		Py_INCREF(nodes[i]->object);
		PyList_Append(list, nodes[i]->object);
	}

	return list;
}

void PyGraph_Free(PyObject *self) {

	D_FREE(self, "Graph");

}

static PyMethodDef PyGraph_methods[] = {
	{"addNode", (PyCFunction)PyGraph_AddNode, METH_VARARGS,
		"add a node to the graph"},
	{"addConnection", (PyCFunction)PyGraph_AddConnection, METH_VARARGS,
		"add a connection to the graph"},
	{"getNumGroups", (PyCFunction)PyGraph_GetNumGroups, METH_VARARGS,
		"add a connection to the graph"},
	{"getGroupNodes", (PyCFunction)PyGraph_GetGroupNodes, METH_VARARGS,
		"add a connection to the graph"},
	{NULL, NULL, 0, NULL}
};

static PyObject *PyGraph_GetAttr(PyObject *self, char *name) {
	return Py_FindMethod(PyGraph_methods, self, name);
}

TOGRA_DEFINE_GETATTR_OBJECT(Graph);
