/*
 * Copyright 2003, Shane Stephens.
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

#include "TograIncludes.h"

#if USE_GLUT
#include <GL/glut.h>
#else
#include <gtk/gtk.h>
#include <gtk/gtkgl.h>
#endif

staticforward PyTypeObject PyRenderControlType;

PyMaterial *current_colour = NULL;
#if !USE_GLUT
static GtkWidget *window;
static GtkWidget *gl_area;
static guint timeout_id = 0;

/* In ms. Aim for 100 fps */
#define TOGRA_TIMEOUT 10

inline gboolean gl_make_current(void);

inline gboolean 
gl_make_current() 
{
  GdkGLContext *glcontext = gtk_widget_get_gl_context (gl_area);
  GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (gl_area);

  return gdk_gl_drawable_make_current (gldrawable, glcontext);
}

static gboolean
timeout_handler (GtkWidget *widget)
{
  /* Invalidate the widget and process updates */
  gdk_window_invalidate_rect (widget->window, &widget->allocation, FALSE);
  gdk_window_process_updates (widget->window, FALSE);

  return TRUE;
}

static gboolean
unmap_event (GtkWidget   *widget,
             GdkEventAny *event,
             gpointer     data)
{
  if (timeout_id != 0) {
    g_source_remove (timeout_id);
    timeout_id = 0;
  }
  return TRUE;
}

static gboolean
map_event (GtkWidget   *widget,
           GdkEventAny *event,
           gpointer     data)
{
  if (timeout_id == 0)
    timeout_id = g_timeout_add(TOGRA_TIMEOUT, (GtkFunction) timeout_handler, widget);
  return TRUE;
}

static gboolean
visibility_notify_event (GtkWidget          *widget,
                         GdkEventVisibility *event,
                         gpointer            data)
{
  if (event->state == GDK_VISIBILITY_FULLY_OBSCURED) {
    if (timeout_id != 0) {
      g_source_remove (timeout_id);
      timeout_id = 0;
    }
  }
  else
    if (timeout_id == 0)
      timeout_id = g_timeout_add(TOGRA_TIMEOUT, (GtkFunction) timeout_handler, widget);
  return TRUE;
}

#else
#define gl_make_current() 1
#endif

PyObject *PyAddMixin(PyObject *self, PyObject *args);
void C_StepMixins(TG_List *mixins, int last, int is_paused, int p_time);
#if USE_GLUT
void C_PyRenderControl_InternalStep(void);
#else
gboolean C_PyRenderControl_InternalStep (gpointer data);
#endif


static inline int PyRenderControl_Check(PyObject *v) {

        return v->ob_type == &PyRenderControlType;

}

static inline long PyRenderControl_BaseTime(PyObject *v) {

        return ((PyRenderControl *)v)->baseTime;

}

static inline int PyRenderControl_IsPaused(PyObject *v) {

        return ((PyRenderControl *)v)->isPaused;

}

static inline int PyRenderControl_isTimeBased(PyObject *v) {

        return ((PyRenderControl *)v)->isTimeBased;

}

struct _C_PyRenderControl_LightObj {
        int active;
        float position[4];
        float diffuse[4];
        float specular[4];
        PyObject *transform;
};

#define NUM_LIGHTS      8

#define LIGHT_UNUSED    -1
#define LIGHT_INACTIVE  0
#define LIGHT_ACTIVE    1

inline void _C_PyRenderControl_InitialiseLighting(PyRenderControl *rc) {

        int i;
        
        rc->lights = malloc(NUM_LIGHTS * sizeof(_C_PyRenderControl_LightObj));

        for (i = 0; i < NUM_LIGHTS; i++)
        {
                memset(&rc->lights[i], 0, sizeof(_C_PyRenderControl_LightObj));
                rc->lights[i].active = LIGHT_UNUSED;
                rc->lights[i].transform = NULL;
        }
        
}

PyObject *this_ptr;

#define MAX_CONTACTS	4

#if !USE_GLUT

static gboolean _delete_event(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	return FALSE;
}

static void _destroy(GtkWidget *widget, gpointer data)
{
	gtk_main_quit();
}

gint _reshape(GtkWidget *widget, GdkEventConfigure *event) 
{
	int w = widget->allocation.width;
	int h = widget->allocation.height;

	if (gl_make_current())
	{
	    glViewport (0, 0, w, h);
	}

	return 0;
}	

gint _init_gl_area(GtkWidget *widget);


int getScreenWidth(void)
{
	return gl_area->allocation.width;
}
int getScreenHeight(void)
{
	return gl_area->allocation.height;
}
#else
int getScreenWidth()
{
	return glutGet(GLUT_WINDOW_WIDTH);
}

int getScreenHeight()
{
	return glutGet(GLUT_WINDOW_HEIGHT);
}
#endif

PyObject *PyGetWidth(PyObject *self, PyObject *args)
{
	return PyInt_FromLong((long)getScreenWidth());
}

PyObject *PyGetHeight(PyObject *self, PyObject *args)
{
	return PyInt_FromLong((long)getScreenHeight());
}

PyObject *C_PyRenderControl_New(int isTimeBased, int isFullScreen)
{

#if USE_GLUT
        float lm[4] = {0.5, 0.5, 0.5, 1};
#endif
        /*
        float lp[4] = {-30, 0, 0, 1};
        float ld[4] = {0.8, 0.8, 0.8, 1};
        */

        PyRenderControl *rc
                = PyObject_New(PyRenderControl, &PyRenderControlType);

        D_ALLOC(rc, "RenderControl");
        
#if USE_GLUT
        int args = 2;
        char *argv[] = {"python", "derivative.py", NULL};
#else
	GdkGLConfig *glconfig;
#endif

        rc->isPaused = FALSE;
        if (isTimeBased == TRUE) {
                rc->baseTime = 0; //GetTime();
        } else {
                rc->baseTime = 0;
        }
	rc->lastTime = rc->baseTime;
        rc->pausedAmount = 0;
        rc->pausedTime = rc->baseTime;
        rc->isTimeBased = isTimeBased;
        rc->renderObject = C_PyContainer_New();
        rc->view_from = C_PyVector_FromValues(-30, 0, 0);
        rc->view_to = C_PyVector_FromValues(0, 0, 0);
        rc->view_up = C_PyVector_FromValues(0, 1, 0);
        rc->keyFunction = NULL;
        rc->mouseFunction = NULL;
        rc->motionFunction = NULL;
	rc->messageFunction = NULL;
        this_ptr = (PyObject *)rc;

        Py_INCREF(rc->renderObject);
        Py_INCREF(rc->view_from);
        Py_INCREF(rc->view_to);
        Py_INCREF(rc->view_up);
	
	TG_list_init(&(rc->mixins));

#if USE_GLUT
	glutInit(&args, argv);
        glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH | GLUT_STENCIL);
	glutInitWindowSize(800, 800);
	glutInitWindowPosition(300, 0);
	glutCreateWindow("Welcome to the World of Togra!");
	if (isFullScreen)
		glutFullScreen();
#else
	gtk_init(NULL, NULL);
	gtk_gl_init (NULL, NULL);

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	if (isFullScreen) {
            gtk_window_fullscreen(GTK_WINDOW (window));
	}
        else {
            gtk_window_set_default_size (GTK_WINDOW (window), 800, 800);
        }
				
	g_signal_connect(G_OBJECT(window), "delete_event", 
			G_CALLBACK(_delete_event), NULL);
	g_signal_connect(G_OBJECT(window), "destroy",
			G_CALLBACK(_destroy), NULL);
	gl_area = gtk_drawing_area_new ();

	glconfig = gdk_gl_config_new_by_mode (GDK_GL_MODE_RGBA | GDK_GL_MODE_DEPTH | GDK_GL_MODE_DOUBLE);
	g_assert (glconfig != NULL);

	gtk_widget_set_gl_capability (gl_area, glconfig, NULL, TRUE, GDK_GL_RGBA_TYPE);

 	gtk_widget_set_events(GTK_WIDGET(gl_area),
                        GDK_EXPOSURE_MASK|
                        GDK_BUTTON_PRESS_MASK|
			GDK_BUTTON_RELEASE_MASK|
			GDK_POINTER_MOTION_MASK|
                        GDK_POINTER_MOTION_HINT_MASK|
			GDK_KEY_PRESS_MASK);	
	
	gtk_container_add(GTK_CONTAINER(window), gl_area);

  	g_signal_connect_after (GTK_OBJECT(gl_area), "realize",
                      GTK_SIGNAL_FUNC(_init_gl_area), NULL);	
	g_signal_connect (GTK_OBJECT(gl_area), "configure_event",
			GTK_SIGNAL_FUNC(_reshape), NULL);
	g_signal_connect (GTK_OBJECT(gl_area), "expose_event",
			GTK_SIGNAL_FUNC(C_PyRenderControl_InternalStep), NULL);
	g_signal_connect (GTK_OBJECT(gl_area), "map_event",
			GTK_SIGNAL_FUNC(map_event), NULL);
	g_signal_connect (GTK_OBJECT(gl_area), "unmap_event",
			GTK_SIGNAL_FUNC(unmap_event), NULL);
	g_signal_connect (GTK_OBJECT(gl_area), "visibility_notify_event",
			GTK_SIGNAL_FUNC(visibility_notify_event), NULL);
	
	gtk_widget_show (window);
	gtk_widget_show (gl_area);

        return (PyObject *)rc;
}

gint _init_gl_area(GtkWidget *widget)
{

        float lm[4] = {0.5, 0.5, 0.5, 1};
	PyRenderControl *rc = (PyRenderControl *)this_ptr;

	if (!gl_make_current())
	{
		printf("Couldn't make GL area current on realize!\n");
		exit(1);
	}

#endif	
	glClearColor(1.0, 1.0, 1.0, 0.0);
        glColor3f(0, 0, 0);
        glPointSize(4);
        glViewport(0, 0, 480, 480);
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        gluPerspective(60, 1, 1.5, 1000);

        glEnable(GL_DEPTH_TEST);
        glEnable(GL_LIGHTING);
        glEnable(GL_NORMALIZE);
        glEnable(GL_CULL_FACE);
	//glEnable(GL_POLYGON_SMOOTH);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glMatrixMode(GL_MODELVIEW);
#if 0
        glEnable(GL_POLYGON_SMOOTH);
#endif
        glLoadIdentity();

        /*
        glLightfv(GL_LIGHT0, GL_POSITION, lp);
        glLightfv(GL_LIGHT0, GL_DIFFUSE, ld);
        glEnable(GL_LIGHT0);

        glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lm);
        */
        glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lm);
	//glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);

        _C_PyRenderControl_InitialiseLighting(rc);

	
#if USE_GLUT	
        return (PyObject *)rc;
#else
	return TRUE;
#endif
}

void C_PyRenderControl_SetAmbientLight(float r, float g, float b, float a) {
	
	float amb[4] = {r, g, b, a};
#if !USE_GLUT
	gl_make_current();
#endif
	
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, amb);

}

PyObject *PyRenderControl_SetAmbientLight(PyObject *self, PyObject *args) {

	float r, g, b, a;
	
	if (!PyArg_ParseTuple(args, "ffff", &r, &g, &b, &a))
		return NULL;

	C_PyRenderControl_SetAmbientLight(r, g, b, a);

	Py_INCREF(Py_None);
	return Py_None;
}

PyObject *PyRenderControl_New(PyObject *self, PyObject *args) {

        int isTimeBased = 1;
	int isFullScreen = 0;

        if (!PyArg_ParseTuple(args, "|ii", &isFullScreen, &isTimeBased)) {
                return NULL;
        }

        return C_PyRenderControl_New(isTimeBased, isFullScreen);
}


#if USE_GLUT
static void C_PyRenderControl_KBFunc(unsigned char key, int x, int y) 
{
	
        PyRenderControl *rc = (PyRenderControl *)this_ptr;

        PyObject *tup = PyTuple_New(3);
        PyTuple_SetItem(tup, 0, PyInt_FromLong(key));
        PyTuple_SetItem(tup, 1, PyInt_FromLong(x));
        PyTuple_SetItem(tup, 2, PyInt_FromLong(y));
	
        PyObject_CallObject(rc->keyFunction, tup);

	return;
}

static void C_PyRenderControl_MouseFunc(int button, int state, int x, int y) 
{
        PyRenderControl *rc = (PyRenderControl *)this_ptr;
        PyObject *tup = PyTuple_New(5);

        PyTuple_SetItem(tup, 0, PyInt_FromLong(button));
        if (state == GLUT_UP)
          PyTuple_SetItem(tup, 1, PyInt_FromLong(0));
        else
          PyTuple_SetItem(tup, 1, PyInt_FromLong(1));
        PyTuple_SetItem(tup, 2, PyInt_FromLong(x));
        PyTuple_SetItem(tup, 3, PyInt_FromLong(y));
	PyTuple_SetItem(tup, 4, PyInt_FromLong(glutGetModifiers()));

        PyObject_CallObject(rc->mouseFunction, tup);
}

static void C_PyRenderControl_MotionFunc(int x, int y) {
        PyRenderControl *rc = (PyRenderControl *)this_ptr;
        PyObject *tup = PyTuple_New(2);

        PyTuple_SetItem(tup, 0, PyInt_FromLong(x));
        PyTuple_SetItem(tup, 1, PyInt_FromLong(y));

        if (PyObject_CallObject(rc->motionFunction, tup) == NULL)
		PyErr_Print();
}

#else

static gboolean C_PyRenderControl_KBFunc(GtkWidget *widget, GdkEventKey *e,
		gpointer data)
{
	
        PyRenderControl *rc = (PyRenderControl *)data;

        PyObject *tup = PyTuple_New(2);
	PyTuple_SetItem(tup, 0, PyInt_FromLong(e->keyval));
	PyTuple_SetItem(tup, 1, PyInt_FromLong(e->state));
	
        PyObject_CallObject(rc->keyFunction, tup);

	return 0;
}

static void C_PyRenderControl_MouseFunc(GtkWidget *w, GdkEventButton *e, 
	gpointer d) 
{
        PyRenderControl *rc = (PyRenderControl *)d;
        PyObject *tup = PyTuple_New(5);

	PyTuple_SetItem(tup, 0, PyInt_FromLong(e->button));
        if (e->type == GDK_BUTTON_RELEASE)
	  PyTuple_SetItem(tup, 1, PyInt_FromLong(0));
        else
	  PyTuple_SetItem(tup, 1, PyInt_FromLong(1));

	PyTuple_SetItem(tup, 2, PyInt_FromLong(e->x));
	PyTuple_SetItem(tup, 3, PyInt_FromLong(e->y));
	PyTuple_SetItem(tup, 4, PyInt_FromLong(e->state));

        PyObject_CallObject(rc->mouseFunction, tup);
}

static void 
C_PyRenderControl_MotionFunc
    (GtkWidget *widget, GdkEventMotion *event, gpointer user_data)
{
  PyRenderControl *rc = (PyRenderControl *)user_data;
  PyObject *tup = PyTuple_New(2);
  gdouble x, y;

  if (event->is_hint) {
    gint i_x, i_y;
    GdkModifierType state;
    gdk_window_get_pointer (widget->window, &i_x, &i_y, &state);
    x = i_x;
    y = i_y;
  } else {
    x = event->x;
    y = event->y;
  }
  PyTuple_SetItem(tup, 0, PyInt_FromLong(x));
  PyTuple_SetItem(tup, 1, PyInt_FromLong(y));

  if (PyObject_CallObject(rc->motionFunction, tup) == NULL)
    PyErr_Print();
}
#endif

inline void _C_PyRenderControl_DoLighting(PyRenderControl *rc, int time) {

        int i;

        for (i = 0; i < NUM_LIGHTS; i++) {
                if (rc->lights[i].active == LIGHT_ACTIVE) {
                        glPushMatrix();
                        if (rc->lights[i].transform != NULL) {
                                GENERATE(rc->lights[i].transform, NULL, time,
                                                1, 0, 1);
                        }
                        glLightfv(GL_LIGHT0 + i, GL_POSITION, 
                                        rc->lights[i].position);
                        glLightfv(GL_LIGHT0 + i, GL_DIFFUSE, 
                                        rc->lights[i].diffuse);
                        glLightfv(GL_LIGHT0 + i, GL_SPECULAR, 
                                        rc->lights[i].specular);
			glPopMatrix();
#if DEBUG_VERBOSE
			printf("LIGHT %d\n", i);
			PRINT_FV("POSITION", rc->lights[i].position);
			PRINT_FV("DIFFUSE", rc->lights[i].diffuse);
			PRINT_FV("SPECULAR", rc->lights[i].specular);
#endif
			
		}
        }

}

void C_PyRenderControl_DoLighting(int time) {

	PyRenderControl *rc = (PyRenderControl *)this_ptr;
	_C_PyRenderControl_DoLighting(rc, time);

}

void C_PyRenderControl_SetupLightingTransform(void) {

        PyRenderControl *rc = (PyRenderControl *)this_ptr;
        double *from, *to, *up;
	
	from = ((PyVector *)(rc->view_from))->vector;
        to = ((PyVector *)(rc->view_to))->vector;
        up = ((PyVector *)(rc->view_up))->vector;
	glLoadIdentity();
        gluLookAt(from[0], from[1], from[2], to[0], to[1], to[2], up[0],
                        up[1], up[2]);

}

int C_PyRenderControl_NewLight(void) {

        int i;
        PyRenderControl *rc = (PyRenderControl *)this_ptr;

        for (i = 0; i < NUM_LIGHTS; i++) {
                if (rc->lights[i].active == LIGHT_UNUSED)
                        break;
        }

        if (i == NUM_LIGHTS)
                return -1;

        rc->lights[i].active = LIGHT_ACTIVE;
	rc->lights[i].transform = NULL;

	gl_make_current();

	glEnable(GL_LIGHT0 + i);
        return i;

}

int C_PyRenderControl_DeleteLight(int light) {

	PyRenderControl *rc = (PyRenderControl *)this_ptr;

	if (rc->lights[light].active == LIGHT_UNUSED)
		return 0;

	rc->lights[light].active = LIGHT_UNUSED;
	gl_make_current();

	glDisable(GL_LIGHT0 + light);

	return 1;
	
}

PyObject *PyRenderControl_DeleteLight(PyObject *self, PyObject *args)
{
	int light;
	
	if (!PyArg_ParseTuple(args, "i", &light))
		return NULL;

	if (!C_PyRenderControl_DeleteLight(light))
		return NULL;

	Py_INCREF(Py_None);
	return Py_None;
}

void C_PyRenderControl_SetLightPosition(int lightId, float *position) {

        PyRenderControl *rc = (PyRenderControl *)this_ptr;

        memcpy(rc->lights[lightId].position, position, 4 * sizeof(float));

}

void C_PyRenderControl_SetLightDiffuse(int lightId, float *diffuse) {

        PyRenderControl *rc = (PyRenderControl *)this_ptr;

        memcpy(rc->lights[lightId].diffuse, diffuse, 4 * sizeof(float));

}

void C_PyRenderControl_SetLightTransform(int lightId, PyObject *transform) {

        PyRenderControl *rc = (PyRenderControl *)this_ptr;

        if (rc->lights[lightId].transform != NULL) {
                Py_DECREF(rc->lights[lightId].transform);
        }

        if (transform != NULL) {
                Py_INCREF(transform);
        }
        rc->lights[lightId].transform = transform;

}

void C_PyRenderControl_SetLightSpecular(int lightId, float *specular) {

	PyRenderControl *rc = (PyRenderControl *)this_ptr;

	memcpy(rc->lights[lightId].specular, specular, 4 * sizeof(float));

}

PyObject *PyRenderControl_NewLight(PyObject *this, PyObject *args) {
        
        PyObject *position, *diffuse = NULL, *transform = NULL, 
		 *specular = NULL;
        float pos[4], dif[4], spec[4];
        int i;
        double *vector;
        PyObject *rval;
       
	/*
	 * TODO: Keyword this
	 */
	
        if (!PyArg_ParseTuple(args, "OO|OO", &position, &diffuse, &transform,
				&specular))
                return NULL;

        vector = ((PyVector *)position)->vector;
        
        for (i = 0; i < 3; i++)
                pos[i] = vector[i];

        /*
         * Eventually, use a flag for this
         */

        pos[3] = 1;

        for (i = 0; i < 4; i++)
                dif[i] = (float)PyFloat_AsDouble(PyTuple_GetItem(diffuse, i));

	if (specular != NULL)
	{
		for (i = 0; i < 4; i++)
			spec[i] = (float)PyFloat_AsDouble(
					PyTuple_GetItem(specular, i));
	}
	else
	{
		spec[0] = 0.5;
		spec[1] = 0.5;
		spec[2] = 0.5;
		spec[3] = 1.0;
	}
	
        if (PyErr_Occurred())
        {
                return NULL;
        }
        
        i = C_PyRenderControl_NewLight();
        if (i == -1)
                return NULL;

        C_PyRenderControl_SetLightPosition(i, pos);
        C_PyRenderControl_SetLightDiffuse(i, dif);
        C_PyRenderControl_SetLightSpecular(i, spec);
        C_PyRenderControl_SetLightTransform(i, transform);

        rval = PyInt_FromLong(i);

        return rval;
        
}

MsgQueue	*std_queue = NULL;

#if USE_GLUT
void C_PyRenderControl_InternalStep(void) {
#else
gboolean C_PyRenderControl_InternalStep (gpointer data) {
#endif


        double *from, *to, *up;
        long time;
        PyRenderControl *rc = (PyRenderControl *)this_ptr;
#if DEBUG_VERBOSE
	printf("Starting frame\n");
#endif
        from = ((PyVector *)(rc->view_from))->vector;
        to = ((PyVector *)(rc->view_to))->vector;
        up = ((PyVector *)(rc->view_up))->vector;

	gl_make_current();

        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
        gluLookAt(from[0], from[1], from[2], to[0], to[1], to[2], up[0],
                        up[1], up[2]);

        time = GetTime() - rc->baseTime - rc->pausedAmount;
	rc->lastTime = time;

        if (rc->isPaused) {
		_C_PyRenderControl_DoLighting(rc, rc->pausedTime - 
				rc->baseTime - rc->pausedAmount);
                RENDER(rc->renderObject, rc->pausedTime - rc->baseTime -
                                rc->pausedAmount, 0);
        } else {
		_C_PyRenderControl_DoLighting(rc, time);
                RENDER(rc->renderObject, time, 0);
        }
	
	
	C_StepMixins(rc->mixins, time, rc->isPaused, rc->pausedAmount);

        Py_BEGIN_ALLOW_THREADS
#if USE_GLUT
        glutSwapBuffers();
        glutPostRedisplay();
#else
	{
	    GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable (gl_area);
	    if (gdk_gl_drawable_is_double_buffered (gldrawable))
	      gdk_gl_drawable_swap_buffers (gldrawable);
	    else
	      glFlush();
	}
#endif
        Py_END_ALLOW_THREADS
		
	SendMessages(std_queue, rc->messageFunction);	
	
/*        
	if (getchar() == 'q')
		exit(0);
*/	
#if !USE_GLUT
	return TRUE;
#endif
}


#define COLOUR_STACK_SIZE       1024
#define DEFAULT_R               0.7
#define DEFAULT_G               0.7
#define DEFAULT_B               0.7
#define DEFAULT_A               1


void C_PyRenderControl_Run(PyObject *this) {

        PyMaterial *default_colour = C_PyMaterial_New();

	current_colour = default_colour;
	
	PyRenderControl *rc = (PyRenderControl *)this;

#if USE_GLUT
        glutDisplayFunc(C_PyRenderControl_InternalStep);
#endif

	/*
        colour_stack = stack_alloc(COLOUR_STACK_SIZE);
        stack_push(colour_stack, default_colour, sizeof(PyMaterial *));
	*/

	rc->baseTime = GetTime();
	rc->lastTime = rc->baseTime;

        COMPILE(rc->renderObject);

	std_queue = NewQueue();

#if USE_GLUT	
        glutMainLoop();
#else
	gtk_main();
#endif
}

static PyObject *PyRenderControl_Run(PyObject *this, PyObject *args)
{
        C_PyRenderControl_Run(this);

        Py_INCREF(Py_None);
        return Py_None;

}

void C_PyRenderControl_Pause(PyObject *this) {

        PyRenderControl *rc = (PyRenderControl *)this;

        rc->isPaused = TRUE;
        rc->pausedTime = GetTime();
}

static PyObject *PyRenderControl_Pause(PyObject *this, PyObject *args) {

        C_PyRenderControl_Pause(this);

        Py_INCREF(Py_None);
        return Py_None;

}

void C_PyRenderControl_Continue(PyObject *this) {

        PyRenderControl *rc = (PyRenderControl *)this;

        rc->isPaused = FALSE;
        rc->pausedAmount += GetTime() - rc->pausedTime;
}

static PyObject *PyRenderControl_Continue(PyObject *this, PyObject *args) {

        C_PyRenderControl_Continue(this);

        Py_INCREF(Py_None);
        return Py_None;

}

static void C_PyRenderControl_Stop(PyObject *this) {

        PyRenderControl *rc = (PyRenderControl *)this;

#if USE_GLUT
        glutIdleFunc(NULL);
#endif
        rc->isPaused = TRUE;
}

static PyObject *PyRenderControl_Stop(PyObject *this, PyObject *args) {

        C_PyRenderControl_Stop(this);

        exit(0);

        Py_INCREF(Py_None);
        return Py_None;

}

static void C_PyRenderControl_SetBackColour(PyObject *this, float r, float g,
                float b)
{
	gl_make_current ();
        glClearColor(r, g, b, 1);
}

static PyObject *PyRenderControl_SetBackColour(PyObject *this, PyObject *args) {

        float r, g, b;

        if (!PyArg_ParseTuple(args, "fff", &r, &g, &b))
                return NULL;

        C_PyRenderControl_SetBackColour(this, r, g, b);

        Py_INCREF(Py_None);
        return Py_None;
}

static int C_PyRenderControl_GetTime(PyObject *this) {

	PyRenderControl *rc = (PyRenderControl *)this;
        return GetTime() - rc->baseTime - rc->pausedAmount;
	
}

static PyObject *PyRenderControl_GetTime(PyObject *this, PyObject *args) {

	int time = C_PyRenderControl_GetTime(this);
	return PyInt_FromLong((long)time);
	
}

static int C_PyRenderControl_Pick(PyObject *this, int *buf, int buf_siz, 
			int x, int y) {

	int port[4];
	double *from, *to, *up;
	PyRenderControl *rc = (PyRenderControl *)this;
        
	from = ((PyVector *)(rc->view_from))->vector;
        to = ((PyVector *)(rc->view_to))->vector;
        up = ((PyVector *)(rc->view_up))->vector;

	gl_make_current();

	glGetIntegerv(GL_VIEWPORT, port);

	glSelectBuffer(buf_siz, buf);
	glRenderMode(GL_SELECT);

	glInitNames();
	glPushName(0);

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluPickMatrix(x, port[3] - y, 1, 1, port);
        gluPerspective(60, 1, 1.5, 1000);

        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
        gluLookAt(from[0], from[1], from[2], to[0], to[1], to[2], up[0],
                        up[1], up[2]);
	
	RENDER(rc->renderObject, rc->lastTime, 1);

	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	return glRenderMode(GL_RENDER);
}

#define PICK_SIZE	256

static PyObject *PyRenderControl_Pick(PyObject *this, PyObject *args) {

	int x, y, i;
	int p_siz = 0;
	PyObject *r;
	int picked[PICK_SIZE];
	int *ptr = picked;
	
	if (!PyArg_ParseTuple(args, "ii", &x, &y)) {
		return NULL;
	}


	p_siz = C_PyRenderControl_Pick(this, picked, PICK_SIZE, x, y);

	r = PyList_New(0);
	for (i = 0; i < p_siz; i++) {
		PyObject *data = PyList_New(0);
		PyObject *obj;
		ptr ++;
		PyList_Append(data, PyInt_FromLong((long)(*ptr++)));
		PyList_Append(data, PyInt_FromLong((long)(*ptr++)));
		obj = (PyObject *)(*ptr++);
		if (obj == NULL) {
			Py_INCREF(Py_None);
			PyList_Append(data, Py_None);
		} else {
			Py_INCREF(obj);
			PyList_Append(data, obj);
		}
		PyList_Append(r, data);
	}

	return r;

}

static PyMethodDef PyRenderControl_methods[] = {
        {"run", (PyCFunction)PyRenderControl_Run, METH_VARARGS,
                "Start rendering the scene graph installed in this controller"},
        {"pause", (PyCFunction)PyRenderControl_Pause, METH_VARARGS,
                "Pause timer in this controller but continue rendering"},
        {"cont", (PyCFunction)PyRenderControl_Continue, METH_VARARGS,
                "Restart timer in this controller"},
        {"exit", (PyCFunction)PyRenderControl_Stop, METH_VARARGS,
                "Stop rendering the scene and exit"},
        {"setBackgroundColour", (PyCFunction)PyRenderControl_SetBackColour,
                METH_VARARGS, "Change the scene background colour"},
        {"newLight", (PyCFunction)PyRenderControl_NewLight,
                METH_VARARGS, "Create a new light"},
        {"deleteLight", (PyCFunction)PyRenderControl_DeleteLight,
                METH_VARARGS, "Delete an existing light"},
	{"setAmbientLight", (PyCFunction)PyRenderControl_SetAmbientLight,
		METH_VARARGS, "change the scene ambient light"},
	{"getTime", (PyCFunction)PyRenderControl_GetTime,
		METH_VARARGS, "get the current render time"},
	{"addMixin", (PyCFunction)PyAddMixin, METH_VARARGS, 
		"add a new mixin"},
	{"pick", (PyCFunction)PyRenderControl_Pick,
		METH_VARARGS, "Find the objects (if any) at a coordinate"},
        {NULL, NULL, 0, NULL}
};

static PyObject *PyRenderControl_GetAttr(PyObject *self, char *name) {

        PyRenderControl *rc = (PyRenderControl *)self;

        if (strcmp(name, "baseContainer") == 0)
        {
                Py_INCREF(rc->renderObject);
                return rc->renderObject;
        }
	else if (strcmp(name, "view_from") == 0)
	{
		Py_INCREF(rc->view_from);
		return rc->view_from;
	}
	else if (strcmp(name, "view_to") == 0)
	{
		Py_INCREF(rc->view_to);
		return rc->view_to;
	}

        return Py_FindMethod(PyRenderControl_methods, self, name);

}

static int PyRenderControl_SetAttr(PyObject *self, char *name,
                PyObject *item)
{
        PyRenderControl *rc = (PyRenderControl *)self;

        if (strcmp(name, "view_from") == 0) {
                Py_DECREF(rc->view_from);
                Py_INCREF(item);
                rc->view_from = item;
        } else if (strcmp(name, "view_to") == 0) {
                Py_DECREF(rc->view_to);
                Py_INCREF(item);
                rc->view_to = item;
        } else if (strcmp(name, "view_up") == 0) {
                Py_DECREF(rc->view_up);
                Py_INCREF(item);
                rc->view_up = item;
        } else if (strcmp(name, "keyFunction") == 0) {
                if (rc->keyFunction != NULL) {
                        Py_DECREF(rc->keyFunction);
		}
                else {
#if USE_GLUT
                        glutKeyboardFunc(C_PyRenderControl_KBFunc);
#else
			
			g_signal_connect(G_OBJECT(window), "key-press-event", 
				G_CALLBACK(C_PyRenderControl_KBFunc), rc);
#endif
		}
                Py_INCREF(item);
                rc->keyFunction = item;
        } else if (strcmp(name, "mouseFunction") == 0) {
                if (rc->mouseFunction != NULL) {
                        Py_DECREF(rc->mouseFunction);
		}
                else {
#if USE_GLUT
                        glutMouseFunc(C_PyRenderControl_MouseFunc);
#else
			g_signal_connect(G_OBJECT(window), "button-press-event",
				G_CALLBACK(C_PyRenderControl_MouseFunc), rc);
			g_signal_connect(G_OBJECT(window), "button-release-event",
				G_CALLBACK(C_PyRenderControl_MouseFunc), rc);
#endif
		}
                Py_INCREF(item);
                rc->mouseFunction = item;
        } else if (strcmp(name, "motionFunction") == 0) {
                if (rc->motionFunction != NULL) {
                        Py_DECREF(rc->motionFunction);
		}
                else {
#if USE_GLUT
                        glutMotionFunc(C_PyRenderControl_MotionFunc);
#else
                        g_signal_connect(G_OBJECT(window), "motion-notify-event",
                            G_CALLBACK(C_PyRenderControl_MotionFunc), rc);
#endif
		}
                Py_INCREF(item);
                rc->motionFunction = item;
        } else if (strcmp(name, "messageFunction") == 0) {
		if (rc->messageFunction != NULL) {
			Py_DECREF(rc->messageFunction);
		}
		if (item == Py_None) {
			rc->messageFunction = NULL;
		} else {
			Py_INCREF(item);
			rc->messageFunction = item;
		}
	} else {
                return 1;
        }

        return 0;

}

static void PyRenderControl_Free(PyObject *self) {

        D_FREE(self, "RenderControl");

}

static PyTypeObject PyRenderControlType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PyRenderControl",
        sizeof(PyRenderControl),
        0,
        PyRenderControl_Free,
        0, /* print */
        PyRenderControl_GetAttr, /* getattr */
        PyRenderControl_SetAttr, /* setattr */
        0, /* compare */
        0, /* repr */
        0, /* as_number */
        0, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        0, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};
