/*
 * Copyright 2003, Shane Stephens.  
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */


/** @file Vector.c 
 * @brief implementation of the PyVector type 
 *
 * This type implements double-precision vectors of length 3 \a only.  It is
 * intended for fast GL applications that use Python as a glue language.  Both
 * Python and C functions for many operations are provided.
 *
 * Following the traditional Python type style, all inputs to the functions
 * provided are of type PyObject.  A utility function, PyVector_Check, is
 * provided to determine whether a PyObject is a PyVector
 *
 * To allow polymorphism, the C bindings all accept a PyObject * PyVector
 * parameter.  This provides a consistent type signature for rendering,
 * indexing, etc.
 */

#include "TograIncludes.h"
staticforward PyTypeObject PyVectorType;

static PyObject *C_PyVector_Add(PyObject *a, PyObject *b);
static PyObject *C_PyVector_MultiplyScalar(PyObject *vector, double s);
static void C_PyVector_TranslateByMe(PyObject *self);
static void C_PyVector_RotateByMe(PyObject *self, double amt);
static IVector PyVector_IVector = {C_PyVector_Add, C_PyVector_MultiplyScalar,
								   C_PyVector_TranslateByMe, 
								   C_PyVector_RotateByMe};
static IC PyVector_Interfaces = {NULL, NULL, NULL, NULL, &PyVector_IVector};

/**
 * Checks if a PyObject is a PyVector
 * @param v The PyObject to check
 * @return TRUE if v is a PyVector
 */
static inline int PyVector_Check(PyObject *v) {

        return (v->ob_type == &PyVectorType);

}

/**
 * Returns the vector held by a PyVector.
 * @param v The PyVector to check
 * @return A (length 3) array of type double
 * @warning Does not check if PyObject is a PyVector
 */
inline double * PyVector_Vector(PyObject *v) {

        return ((PyVector *)v)->vector;

}

/**
 * Constructs a new PyVector [C binding]
 * @return A PyVector with x, y and z values initialised to 0
 */
PyObject *C_PyVector_New(void) {

        PyVector *v;
        v = PyObject_New(PyVector, &PyVectorType);
        v->interfaces = &PyVector_Interfaces;

        D_ALLOC(v, "Vector");
        
        v->vector = malloc(3 * sizeof(double));
        v->vector[0] = v->vector[1] = v->vector[2] = 0;
        v->memory_internal = TRUE;

        return (PyObject *)v;
        
}

/**
 * Constructs a new PyVector and initialises its x, y and z values [C binding]
 * @param x, y, z The x, y and z values for the new vector
 * @return A PyVector with the requested return values
 */

PyObject *C_PyVector_FromValues(double x, double y, double z) {

        PyVector *v = (PyVector *)C_PyVector_New();
        v->vector[0] = x;
        v->vector[1] = y;
        v->vector[2] = z;

        //printf("<%.2f,%.2f,%.2f> ", x, y, z);
        
        return (PyObject *)v;
        
}

/**
 * Constructs a new PyVector from a C double vector of length 3 [C binding]
 * @param vector A double array of length 3
 * @return A PyVector
 */
PyObject *C_PyVector_FromVector(double *vector) {

        PyVector *v;
        v = PyObject_New(PyVector, &PyVectorType);
		v->interfaces = &PyVector_Interfaces;
        v->vector = vector;
        v->memory_internal = FALSE;

        return (PyObject *)v;
        
}

/**
 * Frees a PyVector [C binding]
 * @param vector A PyVector
 */
static void C_PyVector_Free(void *vector) {

        PyVector *v = (PyVector *)vector;

        D_FREE(v, "Vector");
        
        if (v->memory_internal == TRUE) {
                free(v->vector);
        }

        PyObject_Del((PyObject *)v);

}

/**
 * Copies a vector's array to a freshly allocated memory region [C binding]
 * @param vector The vector to detach
 * 
 * This function is required to allow a common GL approach - vector lists.
 * Essentially, vector lists are contiguous regions of memory that can be
 * marked as GL data.  Hence, the PyVector type allows vectors to be
 * constructed with existing vector data.  The constructor does not copy this
 * data - instead it merely points to it in-place.  As with standard
 * Python objects, modifying the vector then modifies the list, and
 * vice-versa.
 *
 * This function allows vector lists to be both treated as arrays of
 * PyVectors, \a and as contiguous arrays of C doubles.  When a vector list is
 * freed, every PyVector it contains is detached.
 */
/*
 * TODO: Do we still need this?
static void C_PyVector_Detach(void *vector) {

        PyVector *v = (PyVector *)vector;
        double *new_vector = malloc(3 * sizeof(double));

        if (v->memory_internal == TRUE) {
                return;
        }

        memcpy(new_vector, v->vector, 3 * sizeof(double));
        v->vector = new_vector;

        v->memory_internal = TRUE;

}
 *
 */
/**
 * Creates a new PyVector, optionally from x, y, z values
 * @param self Unused
 * @param args [Tuple] Either () or (PyFloat, PyFloat, PyFloat)
 * @return A new PyVector
 */
PyObject *PyVector_New(PyObject *self, PyObject *args) {

        int size = PyTuple_Size(args);
        double x, y, z;

        if (size != 0 && size != 3) {
                return NULL;
        }

        if (size == 0) {
                return (PyObject *)C_PyVector_New();
        }

        if (!PyArg_ParseTuple(args, "ddd", &x, &y, &z)) {
                return NULL;
        }

        return (PyObject *)C_PyVector_FromValues(x, y, z);
        
}

/**
 * Frees a PyVector
 * @param self The vector to free
 */
static void PyVector_Free(PyObject *self) {

        C_PyVector_Free((void *)self);
        
}

/**
 * Adds two PyVectors together [C binding]
 * @param a, b The vectors to add
 * @return The sum of \a a and \a b: \f$(a_0+b_0,a_1+b_1,a_2+b_2)\f$, as a
 * PyVector
 */
static PyObject *C_PyVector_Add(PyObject *a, PyObject *b) {

        double *c, *d;
        
        PyVector *v = (PyVector *)a, *w = (PyVector *)b;
        c = v->vector;
        d = w->vector;
        
        return (PyObject *)C_PyVector_FromValues(c[0] + d[0], c[1] + d[1],
                        c[2] + d[2]);

}

/**
 * Adds two PyVectors together
 * @param self The first PyVector
 * @param other The second PyVector
 * @return The sum of \a self and \a other
 */
static PyObject *PyVector_Add(PyObject *self, PyObject *other) {

        if (!PyVector_Check(other)) {
                Py_INCREF(Py_NotImplemented);
                return Py_NotImplemented;
        }

        return C_PyVector_Add(self, other);
        
}

PyObject *C_PyVector_Cross(PyObject *a, PyObject *b) {

        double *c = PyVector_Vector(a);
        double *d = PyVector_Vector(b);

        return (PyObject *)C_PyVector_FromValues(c[1] * d[2] - c[2] * d[1],
                        c[2] * d[0] - c[0] * d[2], c[0] * d[1] - c[1] * d[0]);

}

PyObject *PyVector_Cross(PyObject *self, PyObject *args) {

        PyObject *other = NULL;

        if (!PyArg_ParseTuple(args, "O", &other))
                return NULL;

        return C_PyVector_Cross(self, other);
}

double C_PyVector_Length(PyObject *a) {

        double *v = PyVector_Vector(a);

        return sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);

}

PyObject *PyVector_Len(PyObject *self, PyObject *args) {

        return PyFloat_FromDouble(C_PyVector_Length(self));

}

static PyMethodDef PyVector_methods[] = {
        {"cross", (PyCFunction)PyVector_Cross, METH_VARARGS,
                "Calculate the cross product of this and another vector"},
        {"length", (PyCFunction)PyVector_Len, METH_VARARGS,
                "Calculate the length of the vector"},
        {NULL, NULL, 0, NULL}
};

static PyObject *PyVector_GetAttr(PyObject *self, char *name) {

        return Py_FindMethod(PyVector_methods, self, name);

}

/**
 * Subtracts a PyVector from another PyVector [C binding]
 * @param a, b The vectors to subtract
 * @return The difference of \a a and \a b: \f$(a_0-b_0,a_1-b_1,a_2-b_2)\f$,
 * as a PyVector
 */
PyObject *C_PyVector_Subtract(PyObject *a, PyObject *b) {

        double *c, *d;
        
        PyVector *v = (PyVector *)a, *w = (PyVector *)b;
        c = v->vector;
        d = w->vector;
        
        return (PyObject *)C_PyVector_FromValues(c[0] - d[0], c[1] - d[1],
                        c[2] - d[2]);

}

/**
 * Subtracts a PyVector from another PyVector
 * @param self The vector being subtracted from
 * @param other The vector that is subtracted
 * @return The difference of \a self and \a other, as a PyVector
 */
static PyObject *PyVector_Subtract(PyObject *self, PyObject *other) {

        if (!PyVector_Check(other)) {
                Py_INCREF(Py_NotImplemented);
                return Py_NotImplemented;
        }

        return C_PyVector_Subtract(self, other);
        
}

/**
 * Multiplies a PyVector by a scalar amount [C binding]
 * @param vector The PyVector
 * @param s The scalar
 * @return \f$(a_0 \times s, a_1 \times s, a_2 \times s)\f$
 */
static PyObject *C_PyVector_MultiplyScalar(PyObject *vector, double s) {

        double *v = ((PyVector *)vector)->vector;
        PyVector *w = (PyVector *)C_PyVector_FromValues(v[0] * s, 
													v[1] * s, v[2] * s);

        return (PyObject *)w;

}

/**
 * Calculates the dot product of two PyVectors [C binding]
 * @param a,b The two PyVectors
 * @return \f$a_0 \times b_0 + a_1 \times b_1 + a_2 \times b_2\f$
 */
static double C_PyVector_DotProduct(void *a, void *b) {

        double *v = ((PyVector *)a)->vector;
        double *w = ((PyVector *)b)->vector;

        return v[0] * w[0] + v[1] * w[1] + v[2] * w[2];

}

/**
 * Calculates the product of a PyVector and another PyObject
 * @param self The PyVector
 * @param other [Tuple] (PyFloat,) or (PyVector,)
 * @return A scalar-multiplied PyVector or the dot product of \a self and \a
 * other
 */
static PyObject *PyVector_Multiply(PyObject *self, PyObject *other) {

        double scalar;

        if (PyVector_Check(other)) {

                double result = C_PyVector_DotProduct(self, other);
                return PyFloat_FromDouble(result);

        }

        if (PyFloat_Check(other)) {
                scalar = PyFloat_AsDouble(other);
        } else if (PyInt_Check(other)) {
                scalar = (double)PyInt_AsLong(other);
        } else {
                return NULL;
        }

        return C_PyVector_MultiplyScalar(self, scalar);

}

/**
 * Renders the vector to screen as a 3D point, using OpenGL [C binding]
 * @param vector The PyVector to render
 */
static void C_PyVector_Render(void *vector) {

        glVertex3dv(((PyVector *)vector)->vector);

}

/**
 * Renders the vector to the screen, using OpenGL
 * @param self The vector to render
 * @param args [Tuple] () or (PyInt,) - render mode (GL_POINT, etc)
 * @param dict Unused
 * @return Py_None
 */
static PyObject *PyVector_Call(PyObject *self, PyObject *args,
                PyObject *dict) {

        PyObject *p;
        int gl_mode;
        
        if (PyTuple_Size(args) != 1) {
                return NULL;
        }

        p = PyTuple_GetItem(args, 0);

        if (p != Py_None) {

                if (!(PyInt_Check(p))) {
                        return NULL;
                }

                gl_mode = (int)PyInt_AsLong(p);

                glBegin(gl_mode);

        }

        C_PyVector_Render(self);

        if (p != Py_None) {
                glEnd();
        }

        Py_INCREF(Py_None);
        return Py_None;

}               

static void C_PyVector_TranslateByMe(PyObject *self) {
	
	PyVector *v = (PyVector *)self;
	
	glTranslated(v->vector[0], v->vector[1], v->vector[2]);
	
}

static void C_PyVector_RotateByMe(PyObject *self, double amt) {
	
	PyVector *v = (PyVector *)self;
	
	glRotated(amt, v->vector[0], v->vector[1], v->vector[2]);
	
}

/**
 * Accesses the vector as an array [C binding]
 * @param vector The vector
 * @param index The index into the vector 
 * \f$( 0 \Rightarrow x, 1 \Rightarrow y, 2 \Rightarrow z)\f$
 */

static double C_PyVector_Item(void *vector, int index) {

        return ((PyVector *)vector)->vector[index];

}

/**
 * Accesses the vector as an array
 * @param self The PyVector
 * @param index The index into \a self
 * 
 * @see PyVector_AssItem
 * 
 * This function allows PyVector objects to be treated as lists in Python.
 */

static PyObject *PyVector_Item(PyObject *self, int index) {

        if (index < 0 || index > 2) {
                return NULL;
        }
        
        return PyFloat_FromDouble(C_PyVector_Item(self, index));

}

/**
 * Modifies the vector through an array interface [C binding]
 * @param vector The vector
 * @param index The index into \a vector
 * @param val The new value of \a vector at \a index
 */

static void C_PyVector_AssItem(void *vector, int index, double val) {

        ((PyVector *)vector)->vector[index] = val;

}

/**
 * Modifies the vector through an array interface
 * @param self The vector
 * @param index The index into \a vector
 * @param val [PyFloat] or [PyInt] The new value of \a vector at \a index 
 *
 * @see PyVector_Item
 *
 * This function allows PyVector objects to be treated as lists in Python.
 */
static int PyVector_AssItem(PyObject *self, int index, PyObject *val) {

        if (!(PyInt_Check(val) || PyFloat_Check(val))) {
                return 1;
        }

        if (index < 0 || index > 2) {
                return 1;
        }

        C_PyVector_AssItem(self, index, PyFloat_AsDouble(val));
        return 0;

}

/**
 * Returns the length of a vector (always 3)
 */
static int PyVector_Length(PyObject *self) {
        return (int)PyInt_FromLong(3L);
}

/**
 * Prints a representation of a vector [C binding]
 * @param vector The vector
 * @param fp The FILE to print \a vector to
 * @param flags Ignored
 */
void C_PyVector_Print(void *vector, FILE *fp, int flags) {

        double *v = ((PyVector *)vector)->vector;
        fprintf(fp, "Vector:%15f%15f%15f\n", v[0], v[1], v[2]);

}

/**
 * This function is called to display a vector in Python
 */
static int PyVector_Print(PyObject *self, FILE *fp, int flags) {

        C_PyVector_Print(self, fp, flags);
        return 0;

}

static PyNumberMethods PyVector_numberMethods = {
        PyVector_Add, /* add */
        PyVector_Subtract, /* subtract */
        PyVector_Multiply, /* multiply */
        0, /* divide */
        0, /* remainder */
        0, /* divmod */
        0, /* power */
        0, /* negative */
        0, /* positive */
        0, /* absolute */
        0, /* nonzero */
        0, /* invert */
        0, /* lshift */
        0, /* rshift */
        0, /* and */
        0, /* xor */
        0, /* or */
        0, /* coerce */
        0, /* int */
        0, /* long */
        0, /* float */
        0, /* oct */
        0, /* hex */
        0, /* iadd */
        0, /* isub */
        0, /* imul */
        0, /* idiv */
        0, /* irem */
        0, /* ipow */
        0, /* ilshift */
        0, /* irshift */
        0, /* iand */
        0, /* ixor */
        0, /* ior */
};
static PySequenceMethods PyVector_sequenceMethods = {
        PyVector_Length, /* length */
        0, /* concat */
        0, /* repeat */
        PyVector_Item, /* item */
        0, /* slice */
        PyVector_AssItem, /* ass_item */
        0, /* ass_slice */
        0, /* contains */
        0, /* inplace_concat */
        0 /* inplace_repeat */
};
static PyTypeObject PyVectorType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PyVector",
        sizeof(PyVector),
        0,
        PyVector_Free,
        PyVector_Print, /* print */
        PyVector_GetAttr, /* getattr */
        0, /* setattr */
        0, /* compare */
        0, /* repr */
        &PyVector_numberMethods, /* as_number */
        &PyVector_sequenceMethods, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        PyVector_Call, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};
