/*
 * Copyright 2003, Shane Stephens & Jan Schmidt.  
 *
 * This work is protected by the BSD license.  Please see license.txt in the
 * root directory for more information.
 */

#include "TograIncludes.h"
staticforward PyTypeObject PyVectorListType;

static int PyVectorList_Length(PyObject *v);
static double * PyVectorList_Values(PyObject *v);
static void C_PyVectorList_Append(PyObject *vlist, PyObject *vector);
static PyObject *C_PyVectorList_Item(PyObject *self, int index);
static IVectorList PyVectorList_IVectorList = {PyVectorList_Length, 
											   PyVectorList_Values,
											   C_PyVectorList_Append,
											   C_PyVectorList_Item};
static IC PyVectorList_Interfaces = {NULL, NULL, &PyVectorList_IVectorList};

static inline int PyVectorList_Check(PyObject *v) {

        return v->ob_type == &PyVectorListType;

}

static inline PyObject ** PyVectorList_PyVectors(PyObject *v) {

        return ((PyVectorList *)v)->pyVectors;

}

static double * PyVectorList_Values(PyObject *v) {

        return ((PyVectorList *)v)->values;

}

static int PyVectorList_Length(PyObject *v) {

        return ((PyVectorList *)v)->length;

}

static inline int PyVectorList_Alloced(PyObject *v) {

        return ((PyVectorList *)v)->alloced;

}

PyObject *C_PyVectorList_New(void) {

        PyVectorList *v = PyObject_New(PyVectorList, &PyVectorListType);

        D_ALLOC(v, "VectorList");
        
        v->interfaces = &PyVectorList_Interfaces;
        v->alloced = v->length = 0;
        v->pyVectors = NULL;
        v->values = NULL;

        return (PyObject *)v;

}

PyObject *C_PyVectorList_OfLength(int length) {

        PyVectorList *v = (PyVectorList *)C_PyVectorList_New();
        v->alloced = length;
        if (v->alloced > 0) {
                v->pyVectors = malloc(sizeof(PyObject *) * v->alloced);
                v->values = malloc(sizeof(double) * 3 * v->alloced);
        }

        return (PyObject *)v;

}


PyObject *PyVectorList_New(PyObject *self, PyObject *args) {

        PyObject *l;
        int size;
        
        if (PyTuple_Size(args) == 1) {

                l = PyTuple_GetItem(args, 0);
                
                if (!PyInt_Check(l)) {
                        return NULL;
                }

                size = (int)PyInt_AsLong(l);
                return C_PyVectorList_OfLength(size);

        } else if (PyTuple_Size(args) == 0) {

                return C_PyVectorList_New();

        }
        
        return NULL;

}

/* 
 * This gets tricky.  For each vector:
 * decrease reference count
 * IF memory allocated in array
 * THEN malloc new space and copy
 */

static void C_PyVectorList_Free(void *vlist) {

        PyVectorList *v = (PyVectorList *)vlist;
        PyVector *cur;
        int i;

        D_FREE(vlist, "VectorList");
        
        for (i = 0; i < v->length; i++) {
                cur = (PyVector *)(v->pyVectors[i]);
                if (cur->memory_internal == TRUE) {
                        cur->vector = malloc(3 * sizeof(double));
                        memcpy(cur->vector, &(v->values[i*3]), 
                                        3 * sizeof(double));
                }

                Py_DECREF((PyObject *)cur);
        
        }

        free(v->values);
        free(v->pyVectors);
        
        PyObject_Del((PyObject *)v);

}

static void PyVectorList_Free(PyObject *self) {

        C_PyVectorList_Free((void *)self);
        
}

static void C_PyVectorList_Append(PyObject *vlist, PyObject *vector) {

        PyVectorList *v = (PyVectorList *)vlist;
        PyVector *n = (PyVector *)vector;
	
        if (v->alloced <= v->length) {
                if (v->alloced == 0) {
                        v->alloced++;
                } else {
                        v->alloced *= 2;
                }
                v->pyVectors 
                        = realloc(v->pyVectors, v->alloced * sizeof(void *));
                v->values = realloc(v->values, v->alloced * 3 * sizeof(double));
        }
		
	memcpy(&(v->values[3 * v->length]), n->vector, 3 * sizeof(double));

        if (n->memory_internal == TRUE) {
                free(n->vector);
                n->memory_internal = FALSE;
                n->vector = &(v->values[3 * v->length]);
        }

        v->pyVectors[v->length] = (PyObject *)n;

        Py_INCREF(n);
        v->length++;

}

static void C_PyVectorList_Render(void *vector) {

        PyVectorList *v = (PyVectorList *)vector;
        int i;

        for (i = 0; i < v->length; i++) {
                glVertex3dv(&(v->values[3 * i]));
        }

}

static PyObject *PyVectorList_Call(PyObject *self, PyObject *args,
                PyObject *dict) {

        PyObject *p;
        
        if (!(PyTuple_Size(args) == 1)) {
                return NULL;
        }

        p = PyTuple_GetItem(args, 0);
        
        if (p != Py_None) {

                int gl_type;

                if (!(PyInt_Check(p))) {
                        return NULL;
                }

                gl_type = (int)PyInt_AsLong(p);

                glBegin(gl_type);

        }

        C_PyVectorList_Render((void *)self);

        if (p != Py_None) {

                glEnd();

        }

        Py_INCREF(Py_None);
        return Py_None;

}
        
static PyObject *PyVectorList_Append(PyObject *self, PyObject *args) {

        PyObject *v;
	
        if (!PyArg_ParseTuple(args, "O", &v)) {
                return NULL;
        }

		if (!IS_VECTOR(v)) {
				return NULL;
		}
		
        C_PyVectorList_Append((void *)self, (void *)v);

        Py_INCREF(Py_None);
        return Py_None;
}

static PyObject *C_PyVectorList_Clone(PyObject *self) {

        PyVectorList *ovl = (PyVectorList *)self;
        PyVectorList *nvl 
                = (PyVectorList *)C_PyVectorList_OfLength(ovl->length);
        int i;
        double *p = nvl->values;
        PyVector **q = (PyVector **)nvl->pyVectors;

        nvl->length = ovl->length;
        
        memcpy(nvl->values, ovl->values, 3 * sizeof(double) * ovl->length);

        for (i = 0; i < ovl->length; i++)
        {
                (*q++) = (PyVector *)C_PyVector_FromVector(p);
                p += 3;
        }

        return (PyObject *)nvl;

}
        

static PyObject *PyVectorList_Clone(PyObject *self, PyObject *args) {

        return C_PyVectorList_Clone(self);

}

static PyObject *C_PyVectorList_Reverse(PyObject *self)
{

        PyVectorList *vl = (PyVectorList *)self;
        PyVectorList *nvl = 
                (PyVectorList *)C_PyVectorList_OfLength(vl->length);
        int i;

        for (i = 0; i < vl->length; i++) {

                memcpy(nvl->values + 3 * i, 
                       vl->values + 3 * (vl->length - i - 1),
                       3 * sizeof(double));
                
        }

        nvl->length = vl->length;

        for(i = 0; i < nvl->length; i++) {
                nvl->pyVectors[i] = C_PyVector_FromVector(nvl->values + 3 * i);
        }

        return (PyObject *)nvl;
        
        
}

static PyObject *PyVectorList_Reverse(PyObject *self, PyObject *args) {

        return C_PyVectorList_Reverse(self);

}

static void C_PyVectorList_Transform(PyObject *self, double matrix[3][3],
                PyObject *vector) {

        double nx, ny, nz;
        int i;
        PyVectorList *vl = (PyVectorList *)self;
        PyVector *v = (PyVector *)vector;
        double *p = vl->values;
        
        for (i = 0; i < vl->length; i++) {
                nx = p[0]*matrix[0][0] + p[1]*matrix[0][1] + p[2]*matrix[0][2];
                ny = p[0]*matrix[1][0] + p[1]*matrix[1][1] + p[2]*matrix[1][2];
                nz = p[0]*matrix[2][0] + p[1]*matrix[2][1] + p[2]*matrix[2][2];

                p[0] = nx;
                p[1] = ny;
                p[2] = nz;

                p += 3;

        }
        
        p = vl->values;
        
        for (i = 0; i < vl->length; i++) {
                *p++ += v->vector[0];
                *p++ += v->vector[1];
                *p++ += v->vector[2];
        }
}

static PyObject *PyVectorList_Transform(PyObject *self, PyObject *args) {

        PyObject *m = NULL, *v = NULL;
        PyObject *f;
        double matrix[3][3];
        int i, j;

        if (!PyArg_ParseTuple(args, "OO", &m, &v)) {
                return NULL;
        }

        for(i = 0; i < 3; i++) { /* col */
                for (j = 0; j < 3; j++) { /* row */
                        f = PyList_GetItem(m, i);
                        f = PyList_GetItem(f, j);
                        matrix[i][j] = PyFloat_AsDouble(f);
                }
        }

        C_PyVectorList_Transform(self, matrix, v);
        
        Py_INCREF(Py_None);
        return Py_None;

}

static void C_PyVectorList_Insert(PyObject *self, int pos, PyObject *vector) {

        PyVectorList *vl = (PyVectorList *)self;
        PyVector *v = (PyVector *)vector;
        int to_copy, i;

        if (vl->alloced <= vl->length) {
                if (vl->alloced == 0) {
                        vl->alloced++;
                } else {
                        vl->alloced *= 2;
                }
                vl->pyVectors 
                        = realloc(vl->pyVectors, vl->alloced * sizeof(void *));
                vl->values 
                        = realloc(vl->values, vl->alloced * 3 * sizeof(double));
        }

        to_copy = vl->length - pos;

        for(i = to_copy; i > 0; i--) {
                memcpy(vl->pyVectors + pos + i, vl->pyVectors + pos + i - 1,
                                sizeof(void *));
                memcpy(vl->values + 3 * (pos + i), 
                                vl->values + 3 * (pos + i - 1), 
                                sizeof(double) * 3);
        }

/*        
        memcpy(vl->pyVectors + pos + 1, vl->pyVectors + pos, 
                        to_copy * sizeof(void *));
        memcpy(vl->values + 3 * (pos + 1), vl->values + 3 * pos, 
                        to_copy * sizeof(double) * 3);
*/
        
        memcpy(vl->values + 3 * pos, v->vector, 3 * sizeof(double));
        
        if (v->memory_internal == TRUE) {
                free(v->vector);
                v->memory_internal = FALSE;
                v->vector = vl->values + 3 * pos;
        }

        vl->pyVectors[pos] = (PyObject *)v;

        Py_INCREF(v);
        vl->length++;
        
}

static PyObject *PyVectorList_Insert(PyObject *self, PyObject *args) {

        PyObject *v = NULL;
        int pos = 0;

        if (!PyArg_ParseTuple(args, "iO", &pos, &v)) {
                return NULL;
        }

        C_PyVectorList_Insert(self, pos, v);

        Py_INCREF(Py_None);
        return Py_None;
        
}

static PyMethodDef PyVectorList_methods[] = {
        {"append", (PyCFunction)PyVectorList_Append, METH_VARARGS,
                "Appends a vector to the vector list"},
        {"clone", (PyCFunction)PyVectorList_Clone, METH_VARARGS,
                "Create a copy of this vector list"},
        {"transform", (PyCFunction)PyVectorList_Transform, METH_VARARGS,
                "Transform list by specified matrix"},
        {"insert", (PyCFunction)PyVectorList_Insert, METH_VARARGS,
                "Inserts a vector into the vector list"},
        {"reverse", (PyCFunction)PyVectorList_Reverse, METH_VARARGS,
                "Create a reversed version of this list"},
        {NULL, NULL, 0, NULL}
};

static PyObject *PyVectorList_GetAttr(PyObject *self, char *name) {

        return Py_FindMethod(PyVectorList_methods, self, name);

}

static PyObject *C_PyVectorList_Item(PyObject *self, int index) {
	
	PyVectorList *vl = (PyVectorList *)self;
	
	if (index >= vl->length) {
		return NULL;
	}
	
	return (PyObject *)(vl->pyVectors[index]);
	
}

static PyObject *PyVectorList_Item(PyObject *self, int index) {
	
	return C_PyVectorList_Item(self, index);
	
}

static PySequenceMethods PyVectorList_sequenceMethods = {
        0, /* length */
        0, /* concat */
        0, /* repeat */
        PyVectorList_Item, /* item */
        0, /* slice */
        0, /* ass_item */
        0, /* ass_slice */
        0, /* contains */
        0, /* inplace_concat */
        0 /* inplace_repeat */
};

static PyTypeObject PyVectorListType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PyVectorList",
        sizeof(PyVectorList),
        0,
        PyVectorList_Free,
        0, /* print */
        PyVectorList_GetAttr, /* getattr */
        0, /* setattr */
        0, /* compare */
        0, /* repr */
        0, /* as_number */
        &PyVectorList_sequenceMethods, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        PyVectorList_Call, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};

static PyObject *C_PyVectorList_Zip(PyObject *a, PyObject *b, 
		int a_cnt, int a_skip, int b_cnt, int b_skip) {
        PyVectorList *vla = (PyVectorList *)a;
        PyVectorList *vlb = (PyVectorList *)b;
        PyVectorList *result;
        int a_index, b_index;

        result = (PyVectorList *)C_PyVectorList_OfLength(
                                        vla->length + vlb->length);

	a_index = 0;
	b_index = 0;
	while ((a_index < vla->length) || (b_index < vlb->length))
        {
		int last_index;
		
		last_index = (a_index + a_cnt);
		if (last_index > vla->length)
			last_index = vla->length;

		while (a_index < last_index)
		{
			/* Take a_cnt entries from vla */
                	PyVector *tmp = (PyVector *)C_PyVector_FromVector(
                                                &(vla->values[a_index*3]));
                	VL_APPEND((PyObject *)result, (PyObject *)tmp);
			a_index++;
		}
		a_index += a_skip;

		/* Take b_cnt entries from vlb */
		last_index = (b_index + b_cnt);
		if (last_index > vlb->length)
			last_index = vlb->length;

		while (b_index < last_index)
		{
			/* Take b_cnt entries from vlb */
                	PyVector *tmp = (PyVector *)
                            C_PyVector_FromVector(&(vlb->values[b_index*3]));
                	VL_APPEND((PyObject *)result, (PyObject *)tmp);
			b_index++;
		}
		b_index += b_skip;
        }

        return (PyObject *)result;
        
}

PyObject *PyVectorList_Zip(PyObject *self, PyObject *args)
{
        PyObject *a = NULL, *b = NULL;
	int a_cnt=1, b_cnt=1;
	int a_skip=0, b_skip=0;

        if (!PyArg_ParseTuple(args, "OO|iiii", &a, &b, &a_cnt, &a_skip,
			      &b_cnt, &b_skip)) {
                return NULL;
        }

        return C_PyVectorList_Zip(a, b, a_cnt, a_skip, b_cnt, b_skip);

}
