#include "gsttograsink.h"
#include <gst/video/video.h>

#define TOGRA_SINK_DEFAULT_SIZE 256

static GstElementDetails gst_tograsink_details =
GST_ELEMENT_DETAILS (
    "Video Sink",
    "Sink/Video",
    "Video Output to Togra texture",
    "Jan Schmidt <thaytan@mad.scientist.com>");

static GstStaticPadTemplate gst_tograsink_sink_template =
    GST_STATIC_PAD_TEMPLATE ("sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS ( GST_VIDEO_CAPS_RGB )
);

enum
{
  ARG_0,
  ARG_WIDTH,
  ARG_HEIGHT
};

static void gst_tograsink_dispose (GObject *object);
static void gst_tograsink_finalize (GObject *object);
static GstCaps *gst_tograsink_getcaps (GstBaseSink *bsink);
static gboolean gst_tograsink_setcaps (GstBaseSink *bsink, GstCaps * caps);

static GstVideoSinkClass *parent_class = NULL;

static void
gst_tograsink_set_property (GObject * object, guint prop_id,
    const GValue * value, GParamSpec * pspec)
{
  GstTograSink *tograsink;

  g_return_if_fail (GST_IS_TOGRASINK (object));
  tograsink = GST_TOGRASINK (object);

  switch (prop_id) {
    case ARG_WIDTH:
      tograsink->set_width = g_value_get_int (value);
      break;
    case ARG_HEIGHT:
      tograsink->set_height = g_value_get_int (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_tograsink_get_property (GObject * object, guint prop_id,
    GValue * value, GParamSpec * pspec)
{
  GstTograSink *tograsink;

  g_return_if_fail (GST_IS_TOGRASINK (object));
  tograsink = GST_TOGRASINK (object);

  switch (prop_id) {
    case ARG_WIDTH:
      g_value_set_int (value, tograsink->set_width);
      break;
    case ARG_HEIGHT:
      g_value_set_int (value, tograsink->set_height);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_tograsink_get_times (GstBaseSink * bsink, GstBuffer * buf,
    GstClockTime * start, GstClockTime * end)
{
  GstTograSink 	*tograsink = GST_TOGRASINK (bsink);

  if (GST_BUFFER_TIMESTAMP_IS_VALID (buf)) {
    *start = GST_BUFFER_TIMESTAMP (buf);
    if (GST_BUFFER_DURATION_IS_VALID (buf)) {
      *end = *start + GST_BUFFER_DURATION (buf);
    } else {
      if (tograsink->framerate_n > 0) {
        *end = *start + gst_util_uint64_scale_int (GST_SECOND, 
                            tograsink->framerate_d, tograsink->framerate_n);
      }
    }
  }
}

/* Called from basesink at the right time to show a frame */
static GstFlowReturn
gst_tograsink_show (GstBaseSink * bsink, GstBuffer * buf)
{
  GstTograSink 	*tograsink = GST_TOGRASINK (bsink);
  GstClockTime  buf_time;

  if (GST_BUFFER_TIMESTAMP_IS_VALID (buf)) {
    tograsink->time = GST_BUFFER_TIMESTAMP (buf);
  }

  buf_time = tograsink->time;
  
  if (GST_BUFFER_DURATION_IS_VALID (buf)) {
    tograsink->time += GST_BUFFER_DURATION (buf);
  }
  else if (tograsink->framerate_n > 0) {
    tograsink->time += gst_util_uint64_scale_int (GST_SECOND,
                            tograsink->framerate_d, tograsink->framerate_n);
  }

  /* Lock the mutex and update the output buffer */
  g_mutex_lock (tograsink->out_mutex);
  
  if (tograsink->waiting_buffer) {
    gst_buffer_unref (tograsink->waiting_buffer);
  }
  gst_buffer_ref (buf);
  tograsink->waiting_buffer = buf;
  
  tograsink->waiting_height = tograsink->cur_height;
  tograsink->waiting_width = tograsink->cur_width;
  tograsink->waiting_time = buf_time;

  if ((tograsink->cur_height != tograsink->set_height) ||
     (tograsink->cur_width != tograsink->set_width)) {
    GstCaps *new_caps = gst_tograsink_getcaps (GST_BASE_SINK (tograsink));
    gst_pad_set_caps (tograsink->sink_pad, new_caps);
    gst_caps_unref (new_caps);
  }

  g_mutex_unlock (tograsink->out_mutex);    

  return GST_FLOW_OK;
}

#if 0
	if (GST_IS_EVENT (data)) {
    		gst_pad_event_default (pad, GST_EVENT (data));
    		return;
  }
#endif

GstTograSinkBuffer *
gst_tograsink_get_buffer (GstTograSink *tograsink)
{
  /* Check the pointer before bothering to lock */
  if (!tograsink->waiting_buffer)
    return NULL;    

  g_mutex_lock (tograsink->out_mutex);
  if (tograsink->curout_buffer)
    gst_buffer_unref (tograsink->curout_buffer);
    
  tograsink->curout_buffer = tograsink->waiting_buffer;
  tograsink->waiting_buffer = NULL;
  
  tograsink->out_buffer.data = GST_BUFFER_DATA (tograsink->curout_buffer);
  tograsink->out_buffer.time = tograsink->waiting_time;
  tograsink->out_buffer.width = tograsink->waiting_width;
  tograsink->out_buffer.height = tograsink->waiting_height;
	
  g_mutex_unlock (tograsink->out_mutex);

  return &(tograsink->out_buffer);
}

static GstCaps *
gst_tograsink_getcaps (GstBaseSink *bsink)
{
  GstTograSink *tograsink = GST_TOGRASINK (bsink);
  GstCaps *caps;

  caps = gst_caps_copy (
	    gst_pad_get_pad_template_caps (GST_VIDEO_SINK_PAD (tograsink)));

#if 0
  gst_caps_set_simple (caps, "width", GST_TYPE_INT_RANGE, 16, 
	tograsink->set_width, "height", GST_TYPE_INT_RANGE, 16, 
	tograsink->set_height, NULL);
#else
  gst_caps_set_simple (caps, "width", G_TYPE_INT,
	tograsink->set_width, "height", G_TYPE_INT,
	tograsink->set_height, NULL);
#endif

  return caps;
}

static gboolean 
gst_tograsink_setcaps (GstBaseSink *bsink, GstCaps * caps)
{
  GstTograSink *tograsink;
  GstStructure *structure;
  gboolean ret;

  tograsink = GST_TOGRASINK (bsink);
  structure = gst_caps_get_structure (caps, 0);
  ret = gst_structure_get_int (structure, "width",
      &(tograsink->cur_width));
  ret &= gst_structure_get_int (structure, "height",
      &(tograsink->cur_height));
  ret &= gst_structure_get_fraction (structure, "framerate",
      &tograsink->framerate_n, &tograsink->framerate_d);

  if (tograsink->cur_width > tograsink->set_width)
    return FALSE;
	  
  if (tograsink->cur_height > tograsink->set_height)
    return FALSE;

  return ret;
}

static GstStateChangeReturn
gst_tograsink_change_state (GstElement * element, GstStateChange transition)
{
  GstTograSink *tograsink;

  tograsink = GST_TOGRASINK (element);

  switch (transition) {
    case GST_STATE_CHANGE_READY_TO_PAUSED:
    tograsink->time = 0; 
    break;
    default:
    break;
  }

  return GST_ELEMENT_CLASS (parent_class)->change_state (element, transition);
}

static void
gst_tograsink_base_init (gpointer g_class)
{
  GstElementClass *element_class = GST_ELEMENT_CLASS (g_class);

  gst_element_class_set_details (element_class, &gst_tograsink_details);

  gst_element_class_add_pad_template (element_class,
      gst_static_pad_template_get (&gst_tograsink_sink_template));
}

static void
gst_tograsink_class_init (GstTograSinkClass * klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;
  GstBaseSinkClass *gstbasesink_class;

  gobject_class = (GObjectClass *) klass;
  gstelement_class = (GstElementClass *) klass;
  gstbasesink_class = (GstBaseSinkClass *) klass;

  parent_class = g_type_class_ref (GST_TYPE_VIDEO_SINK);

  gobject_class->dispose = gst_tograsink_dispose;
  gobject_class->finalize = gst_tograsink_finalize;
  gobject_class->set_property = gst_tograsink_set_property;
  gobject_class->get_property = gst_tograsink_get_property;

  g_object_class_install_property (gobject_class, ARG_WIDTH,
      g_param_spec_int ("width", "width", 
	      "The maximum width for the output texture", 
	      1, 4096, TOGRA_SINK_DEFAULT_SIZE, G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class, ARG_HEIGHT,
      g_param_spec_int ("height", "height", 
		"The maximum height for the output texture",
          	1, 4096, TOGRA_SINK_DEFAULT_SIZE, G_PARAM_READWRITE));

  gstelement_class->change_state = gst_tograsink_change_state;

  gstbasesink_class->get_caps = GST_DEBUG_FUNCPTR (gst_tograsink_getcaps);
  gstbasesink_class->set_caps = GST_DEBUG_FUNCPTR (gst_tograsink_setcaps);
  gstbasesink_class->get_times = GST_DEBUG_FUNCPTR (gst_tograsink_get_times);
  gstbasesink_class->preroll = GST_DEBUG_FUNCPTR (gst_tograsink_show);
  gstbasesink_class->render = GST_DEBUG_FUNCPTR (gst_tograsink_show);
}


static void
gst_tograsink_init (GstTograSink *tograsink)
{
  tograsink->time = 0;
  tograsink->cur_width = -1;
  tograsink->cur_height = -1;
  tograsink->framerate_n = 0;
  tograsink->framerate_d = 0;
  tograsink->set_width = TOGRA_SINK_DEFAULT_SIZE;
  tograsink->set_height = TOGRA_SINK_DEFAULT_SIZE;
  
  tograsink->out_mutex = g_mutex_new();
  tograsink->waiting_buffer = NULL;
  tograsink->curout_buffer = NULL;
}

static void
gst_tograsink_dispose (GObject *object)
{
  GstTograSink *tograsink = GST_TOGRASINK(object);

  if (tograsink->waiting_buffer) {
    gst_buffer_unref (tograsink->waiting_buffer);
    tograsink->waiting_buffer = NULL;
  }
  if (tograsink->curout_buffer) {
    gst_buffer_unref (tograsink->curout_buffer);
    tograsink->curout_buffer = NULL;
  }
}

static void
gst_tograsink_finalize (GObject *object)
{
  GstTograSink *tograsink = GST_TOGRASINK(object);
  g_mutex_free (tograsink->out_mutex);
}


GType gst_tograsink_get_type (void)
{
	static GType tograsink_type = 0;
	if (G_UNLIKELY (!tograsink_type)) {
    		static const GTypeInfo tograsink_info = {
      				sizeof (GstTograSinkClass),
      				gst_tograsink_base_init,
      				NULL,
      				(GClassInitFunc) gst_tograsink_class_init,
      				NULL,
      				NULL,
      				sizeof (GstTograSink),
      				0,
      				(GInstanceInitFunc) gst_tograsink_init,
    		};
		
    		tograsink_type =
        		g_type_register_static (GST_TYPE_VIDEO_SINK,
			"GstTograSink", &tograsink_info, 0);
	}

	return tograsink_type;
}

