/*
 * Python object that plays a movie into a texture using GStreamer
 */

#include "TograIncludes.h"
#include "PyGstTexture_common.h"

#include <pygobject.h>
#include "gst/gsttograsink.h"

typedef struct {
        PyObject_HEAD
        IC *interfaces;
        uint texture[2];
        int width;
        int height;
	int firstframe;
	GstTograSink *sink;
} PyGstTexture;

static void C_PyGstTexture_SetGstTexture(PyObject *material, int time);

static ITexture PyGstTexture_IGstTexture = 
	{C_PyGstTexture_SetGstTexture};
static IC PyGstTexture_Interfaces = {NULL, NULL, NULL, NULL, NULL, NULL, 
	NULL, &PyGstTexture_IGstTexture};
static int gsttexture_nextframe (PyGstTexture *gst_tex, int time);

static void do_init_pygobject (int *inited)
{
            init_pygobject();
            *inited = 1;
}

PyObject *C_PyGstTexture_New(int w, int h, char *name) {
        PyGstTexture *m = NULL;
        GstTograSink *sink = NULL;
        static int inited = 0;

        if (!inited) {
            do_init_pygobject (&inited);
            if (!inited)
                return NULL;

	    /* We rely on GStreamer being initialised. Import pygst and call
	    * pygst.require */
	    if (!PyRun_SimpleString("import pygst; pygst.require('0.10'); import gst") && 0) {
		PyErr_SetString(PyExc_ImportError, \
		    "Could not import gst-0.10"); \
		return NULL;
	    }
        }

        sink = GST_TOGRASINK (g_object_new (GST_TYPE_TOGRASINK, "name", name, 
		    "width", w, "height", h, NULL));

        g_object_ref (G_OBJECT (sink));

	m = (PyGstTexture *)PyObject_New(PyGstTexture, 
		&PyGstTextureType);
	if (!m) {
		g_object_unref (G_OBJECT (sink));
		return NULL;
	}

	D_ALLOC(m, "GstTexture");

        m->interfaces = &PyGstTexture_Interfaces;
	
	m->width  = w;
	m->height = h;
	m->sink = sink;
	m->firstframe = 1;
        glGenTextures(1, m->texture);
	
        return (PyObject *)m;
}

static int
pygst_powerof2 (int val)
{
	/* Shift the lowest set bit to the bottom */
	while ((val & 0x01) == 0)
		val = val >> 1;
	/* If only one bit is set, it's a power of 2 */
	if (val == 0x01)
		return 1;

	return 0;
}

PyObject *PyGstTexture_New(PyObject *self, PyObject *args) {

        PyObject *t;
        int width, height;
	char *name = NULL;
        
        if (!PyArg_ParseTuple(args, "ii|s", &width, &height, &name))
                return NULL;

	/* Only allow power of 2 for width and height */
	if (!pygst_powerof2 (width))
		return NULL;
	if (!pygst_powerof2 (height))
		return NULL;

        t = C_PyGstTexture_New (width, height, name);

        return t;
}

static void PyGstTexture_Free(PyObject *obj) {
        PyGstTexture *gst_tex = (PyGstTexture *) obj;

        g_object_unref (gst_tex->sink);
        D_FREE(obj, "GstTexture");
}

static int
gsttexture_nextframe (PyGstTexture *gst_tex, int time) {
        GstTograSink *sink = gst_tex->sink;
	unsigned char *frame = NULL;
        GstTograSinkBuffer *buffer;

        buffer = gst_tograsink_get_buffer (sink);
        if (buffer) {
            if (buffer->width != gst_tex->width ||
                buffer->height != gst_tex->height) {
                //g_print("Incoming buffer is wrong size!\n");
                //return 0;
		gst_tex->width = buffer->width;
		gst_tex->height = buffer->height;
                gst_tex->firstframe = 1;
            }

            frame = buffer->data;
        } else if (!gst_tex->firstframe) {
            /* No work to do */
            return 0;
        }

        glBindTexture(GL_TEXTURE_2D, gst_tex->texture[0]);

        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	
        if (!frame) {
            /* If we don't have a frame yet, initalise the texture to black */
            frame = malloc (3 * gst_tex->height *  gst_tex->width);
            if (!frame) {
                PyErr_SetString(PyExc_MemoryError, \
                    "Out of memory clearing texture"); \
                return 0;
            }

            memset (frame, 0, 3 * gst_tex->height *  gst_tex->width);
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, gst_tex->width, 
	       	gst_tex->height, 0, GL_RGB, GL_UNSIGNED_BYTE, (void *)frame);
	    gst_tex->firstframe = 0;

            free (frame);
        }
        else if (gst_tex->firstframe) {
          glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, gst_tex->width, gst_tex->height, 0, GL_RGB,
                          GL_UNSIGNED_BYTE, (void *)frame);
	  gst_tex->firstframe = 0;
	}
	else
	{
#if 0
          glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, gst_tex->width, gst_tex->height, GL_RGB, 
                           GL_UNSIGNED_BYTE, (void *)frame);
#else
	  /*
	  int y;
	  unsigned char *line = frame;
	  int slice_height = 8;
	  int stride = 3 * slice_height * gst_tex->width;

	  
	  for (y = 0; y < gst_tex->height; y += slice_height)
	  { 
          	//glTexSubImage2D(GL_TEXTURE_2D, 0, 0, y, gst_tex->width, slice_height, GL_RGB, 
                //	GL_UNSIGNED_BYTE, (void *)line);
		line += stride;
	  }
	  */
	  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, gst_tex->width,
			gst_tex->height, 0, GL_RGB, GL_UNSIGNED_BYTE,
			(void *)frame);
#endif
	}

	return 1;
}

static void C_PyGstTexture_SetGstTexture(PyObject *m, int time) {
	PyGstTexture *gst_tex = (PyGstTexture *)m;

	glEnable(GL_TEXTURE_2D);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	if (!gsttexture_nextframe (gst_tex, time)) 
		glBindTexture(GL_TEXTURE_2D, ((PyGstTexture *)m)->texture[0]);
}

GstTograSink *C_PyGstTexture_GetSink(PyObject *self) 
{
    return ((PyGstTexture *)self)->sink;
}

PyObject *PyGstTexture_GetSink(PyObject *self, PyObject *args) {
    GstTograSink *sink = C_PyGstTexture_GetSink(self);

    return pygobject_new((GObject *)sink);
}

static PyMethodDef PyGstTexture_Methods[] = {
        {"getSink", (PyCFunction)PyGstTexture_GetSink,
                METH_VARARGS, "Return the TograSink object for use in a GStreamer pipeline"},
        {NULL, NULL, 0, NULL}
};

static PyObject *PyGstTexture_GetAttr(PyObject *self, char *name) {

        return Py_FindMethod(PyGstTexture_Methods, self, name);

}

static PyTypeObject PyGstTextureType = {
        PyObject_HEAD_INIT(NULL)
        0,
        "PyContainer",
        sizeof(PyGstTexture),
        0,
        PyGstTexture_Free,
        0, /* print */
        PyGstTexture_GetAttr, /* getattr */
        0, /* setattr */
        0, /* compare */
        0, /* repr */
        0, /* as_number */
        0, /* as_sequence */
        0, /* as_mapping */
        0, /* hash */
        0, /* call */
        0, /* str */
        0, /* getattro */
        0, /* setattro */
        0, /* as_buffer */
        0, /* flags */
};
