# Testing graph design ideas - this should conform to the
# following interface
#
#	GetNumConnections() -> int
#	GetNextConnection() -> (Node, Node, length)
#	SupportsGroups() -> bool
#	GetNumGroups() -> int
#	GetNumNodes() -> int
#	GetNextNode() -> Node
#	NextGroup()
#	GetGroupNumNodes() -> int
#	GetGroupNumConnections() -> int
#	GetNextGroupConnection() -> (Node, Node, length)
#	GetNextGroupNode() -> Node

class Token:
	def __init__(this, data):
		this.data = data

class Node:
	def __init__(this):
		this.data = None
		this.next = None
	def init(this, data):
		this.data = data
		this.group = None
	def setGroup(this, group):
		this.group = group
	def getGroup(this):
		return this.group
	def __repr__(this):
		return "<NODE %s>" % (str(this.data),)

class Connection:
	def __init__(this):
		this.next = None
	def init(this, node1, node2):
		this.n1 = node1
		this.n2 = node2
		this.group = None
	def setGroup(this, group):
		this.group = group
	def getGroup(this):
		return this.group

# Note: deletion from groups is always going to be messy, so I'm not going
# to try any neat tricks here
class Group:
	def __init__(this):
		this.nodes = []
		this.connections = []
		this.next = None
	def init(this, node):
		this.nodes.append(node)
		node.setGroup(this)
	def merge(this, otherGroup):
		for node in otherGroup.nodes:
			node.setGroup(this)
		for connection in otherGroup.connections:
			connection.setGroup(this)
		this.nodes += otherGroup.nodes
		this.connections += otherGroup.connections
	def addConnection(this, c):
		this.connections.append(c)
		c.setGroup(this)
	def getNextNode(this, token):
		token.data += 1
		try:
			return this.nodes[token.data]
		except:
			return None
	def getNextConnection(this, token):
		token.data += 1
		try:
			return this.connections[token.data]
		except:
			return None
	def startGetNodes(this):
		return Token(-1)
	def startGetNodesFrom(this, f):
		return Token(f.data)
	def startGetConnections(this):
		return Token(-1)
	def getSize(this):
		return len(this.nodes)
	def __repr__(this):
		return "<GROUP %s %s>" % (str(this.nodes), str(this.connections))

class ItemList:
	def __init__(this, type):
		this.type = type
		this.items = None
		this.inactiveItems = None
		this.freeItems = None
		this.numItems = 0
	def addNew(this):
		if this.freeItems is not None:
			f = this.freeItems
			this.freeItems = f.next
			f.next = None
		else:
			f = this.type()
		this.numItems += 1
		f.next = this.items
		this.items = f
		return f
	def delete(this, item):
		if this.items == item:
			this.items = this.items.next
		else:
			p = this.items
			if p is None:
				return
			q = this.items.next
			
			while q is not None:
				if q == item:
					p.next = q.next
					break
				p = q
				q = q.next
				
		item.next = this.freeItems
		this.freeItems = item
		this.numItems -= 1
	def getSize(this):
		return this.numItems
	def startGet(this):
		return Token(this.items)
	def startGetFrom(this, f):
		return Token(f.data)
	def getNext(this, token):
		r = token.data
		try:
			token.data = token.data.next
		except:
			pass
		return r


class Graph:
	def __init__(this):
		this.groups = ItemList(Group)
		this.connections = ItemList(Connection)
		this.nodes = ItemList(Node)

	def AddNode(this, node):
		n = this.nodes.addNew()
		n.init(node)
		
		g = this.groups.addNew()
		g.init(n)

		return n
	
	def AddConnection(this, n1, n2):
		c = this.connections.addNew()
		c.init(n1, n2)

		if not n1.getGroup() == n2.getGroup():
			this.merge(n1.getGroup(), n2.getGroup())
		n1.getGroup().addConnection(c)
		
	def merge(this, g1, g2):
		g1.merge(g2)
		this.groups.delete(g2)
			
	def StartGetConnections(this):
		return this.connections.startGet()
	
	def GetNextConnection(this, token):
		c = this.connections.getNext(token)
		try:
			return c.n1.data, c.n2.data
		except:
			return None

	def StartGetNodes(this, *arg):
		if len(arg) == 1:
			return this.nodes.startGetFrom(arg[0])
		else:
			return this.nodes.startGet()

	def GetNextNode(this, token):
		n = this.nodes.getNext(token)
		try:
			return n.data
		except:
			return None

	def SupportsGroups(this):
		return 1

	def StartGetGroups(this, *arg):
		if len(arg) == 1:
			return this.groups.startGetFrom(arg[0])
		else:
			return this.groups.startGet()

	def NextGroup(this, token):
		return this.groups.getNext(token)

	def StartGetGroupNodes(this, group, *arg):
		if len(arg) == 1:
			return group.startGetNodesFrom(arg[0])
		return group.startGetNodes()

	def StartGetGroupConnections(this, group):
		return group.startGetConnections()
		
	def GetNextGroupConnection(this, group, token):
		c = group.getNextConnection(token)
		try:
			return c.n1.data, c.n2.data
		except:
			return None
	
	def GetNextGroupNode(this, group, token):
		n = group.getNextNode(token)
		try:
			return n.data
		except:
			return None

	def GetGroupSize(this, group):
		return group.getSize()


# unit test

if __name__ == "__main__":
	def dump_graph(g):
		t = g.StartGetNodes()
		while 1:
			z = g.GetNextNode(t)
			if z is None:
				break
			print z
		
		t = g.StartGetConnections()
		while 1:
			z = g.GetNextConnection(t)
			if z is None:
				break
			print z
		
		print "by group"
		t = g.StartGetGroups()
		while 1:
			gg = g.NextGroup(t)
			if gg is None:
				break
			print "new group"
			t1 = g.StartGetGroupNodes(gg)
			t2 = g.StartGetGroupConnections(gg)
			while 1:
				z = g.GetNextGroupNode(gg, t1)
				if z is None:
					break
				print z
			while 1:
				z = g.GetNextGroupConnection(gg, t2)
				if z is None:
					break
				print z
		
	g = Graph()
	a = g.AddNode(1)
	b = g.AddNode(2)
	c = g.AddNode(3)

	g.AddConnection(a, b)
	dump_graph(g)
	g.AddConnection(a, c)
	dump_graph(g)
